from pytest import approx, importorskip
import numpy as np
sh = importorskip("sharrow")
import larch.numba as lx
import larch.numba.data_arrays as lxd
from larch.dim_names import CASEID, ALTID

def test_weighted():
    m = lx.example(1, legacy=True)
    m.choice_ca_var = 'chose'
    m.weight_co_var = 'hhinc+100'
    ds = lxd.to_dataset(m.dataservice)
    y, flows = lxd.prepare_data(ds, m)
    assert isinstance(y, sh.Dataset)
    assert sorted(y.coords.keys()) == sorted([CASEID, 'alt_names', ALTID, 'var_co', 'var_ca'])
    assert list(y.keys()) == ['co', 'ca', 'ch', 'wt', 'av']
    assert y.dims == {CASEID: 5029, ALTID: 6, 'var_co': 1, 'var_ca': 2}
    assert y.wt.values[:3] == approx(np.array([142.5, 117.5, 112.5], dtype=np.float32))


def test_choice_code():
    m = lx.example(1, legacy=True)
    m.choice_co_code = 'chosen_alt'
    m.weight_co_var = 'hhinc+100'
    chosen_alt = np.where(m.dataservice.data_ca['chose'].unstack())[1] + 1
    ds = lxd.to_dataset(m.dataservice)
    ds['chosen_alt'] = sh.DataArray(
        chosen_alt,
        dims=['_caseid_',],
    )
    y, flows = lxd.prepare_data(ds, m)
    assert isinstance(y, sh.Dataset)
    assert sorted(y.coords.keys()) == sorted([CASEID, 'alt_names', ALTID, 'var_co', 'var_ca'])
    assert list(y.keys()) == ['co', 'ca', 'ch', 'wt', 'av']
    assert y.dims == {CASEID: 5029, ALTID: 6, 'var_co': 1, 'var_ca': 2}
    assert y.wt.values[:3] == approx(np.array([142.5, 117.5, 112.5], dtype=np.float32))


def test_shared_data():
    m = lx.example(1, legacy=True)
    m.choice_ca_var = 'chose'
    m.weight_co_var = 'hhinc+100'
    ds = lxd.to_dataset(m.dataservice)
    pool = lx.DataTree(base=ds)
    y, flows = lxd.prepare_data(pool, m)
    assert isinstance(y, sh.Dataset)
    assert sorted(y.coords.keys()) == sorted([CASEID, 'alt_names', ALTID, 'var_co', 'var_ca'])
    assert list(y.keys()) == ['co', 'ca', 'ch', 'wt', 'av']
    assert y.dims == {CASEID: 5029, ALTID: 6, 'var_co': 1, 'var_ca': 2}
    assert y.wt.values[:3] == approx(np.array([142.5, 117.5, 112.5], dtype=np.float32))



def example(extract='m', estimate=False):
    import larch.numba as lx
    from larch import P, X

    hh, pp, tour, skims, emp = lx.example(200, ['hh', 'pp', 'tour', 'skims', 'emp'])

    logsums_file = lx.example(202, output_file='/tmp/logsums.zarr.zip')
    logsums = lx.DataArray.from_zarr('/tmp/logsums.zarr.zip')

    ca = lx.Dataset(
        {'logsum': logsums},
        caseid='TOURID',
        alts=skims.TAZ_ID,
    )
    ca

    emp.info()

    tree = lx.DataTree(
        base=ca,
        tour=tour.rename_axis(index='TOUR_ID'),
        hh=hh.set_index("HHID"),
        person=pp.set_index('PERSONID'),
        emp=emp,
        skims=lx.Dataset.from_omx(skims),
        relationships=(
            "base.TAZ_ID @ emp.TAZ",
            "base.TOURID -> tour.TOUR_ID",
            "tour.HHID @ hh.HHID",
            "tour.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ skims.otaz",
            "base.TAZ_ID @ skims.dtaz"
        ),
    )

    m = lx.Model(datatree=tree)
    m.title = "Exampville Work Tour Destination Choice v1"

    m.quantity_ca = (
            + P.EmpRetail_HighInc * X('RETAIL_EMP * (INCOME>50000)')
            + P.EmpNonRetail_HighInc * X('NONRETAIL_EMP') * X("INCOME>50000")
            + P.EmpRetail_LowInc * X('RETAIL_EMP') * X("INCOME<=50000")
            + P.EmpNonRetail_LowInc * X('NONRETAIL_EMP') * X("INCOME<=50000")
    )

    m.quantity_scale = P.Theta


    m.utility_ca = (
        + P.logsum * X.logsum
        + P.distance * X.AUTO_DIST
    )

    m.choice_co_code = "tour.DTAZ"

    m.lock_values(
        EmpRetail_HighInc=0,
        EmpRetail_LowInc=0,
    )

    m.loglike()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.maximize_loglike()

    m.calculate_parameter_covariance()

    m.distribution_on_idca_variable('AUTO_DIST')

    m.distribution_on_idca_variable('AUTO_DIST', bins=40, range=(0,10))

    m.distribution_on_idca_variable(
        'AUTO_DIST',
        style='kde',
    )

    m.distribution_on_idca_variable(
        'AUTO_DIST',
        xlabel="Distance (miles)",
        bins=26,
        subselector='INCOME<10000',
        range=(0,13),
        header='Destination Distance, Very Low Income (<$10k) Households',
    )

    report = lx.Reporter(title=m.title)

    report << '# Parameter Summary' << m.parameter_summary()

    report << "# Estimation Statistics" << m.estimation_statistics()

    report << "# Utility Functions" << m.utility_functions()

    figure = m.distribution_on_idca_variable(
        'AUTO_DIST', 
        xlabel="Distance (miles)",
        style='kde',
        header='Destination Distance',
    )
    report << "# Visualization"
    report << figure

    report.save(
        '/tmp/exampville_dest_choice.html',
        overwrite=True,
        metadata=m,
    )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import pandas as pd
    import geopandas as gpd
    import larch
    larch.__version__

    import larch.exampville

    larch.exampville.files.shapefile

    taz_shape = gpd.read_file("zip://"+larch.exampville.files.shapefile)

    larch.exampville.files.employment

    emp = pd.read_csv(larch.exampville.files.employment, index_col='TAZ')

    emp.head()

    larch.exampville.files.skims

    skims = larch.OMX( larch.exampville.files.skims, mode='r' )
    skims

    larch.exampville.files.hh

    hh = pd.read_csv( larch.exampville.files.hh )

    hh.head()

    larch.exampville.files.person

    pp = pd.read_csv( larch.exampville.files.person )

    pp.head()

    larch.exampville.files.tour

    tour = pd.read_csv( larch.exampville.files.tour )

    tour.head()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import pandas as pd
    import larch.numba as lx

    raw_data = pd.read_csv(lx.example_file('swissmetro.csv.gz')).rename_axis(index='CASEID')
    data = lx.Dataset.from_idco(raw_data, alts={1:'Train', 2:'SM', 3:'Car'})
    m = lx.Model(data.query_cases("PURPOSE in (1,3) and CHOICE != 0"))

    m.title = "swissmetro example 09 (nested logit)"

    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }
    m.choice_co_code = 'CHOICE'

    from larch.roles import P,X
    m.utility_co[1] = ( P.ASC_TRAIN
                      + P.B_TIME * X.TRAIN_TT
                      + P.B_COST * X("TRAIN_CO*(GA==0)") )
    m.utility_co[2] = ( P.B_TIME * X.SM_TT
                      + P.B_COST * X("SM_CO*(GA==0)") )
    m.utility_co[3] = ( P.ASC_CAR
                      + P.B_TIME * X.CAR_TT
                      + P.B_COST * X("CAR_CO") )

    m.graph.new_node(parameter="existing", children=[1,3], name='Existing')

    m.graph

    m.ordering = [
        ("ASCs", 'ASC.*',),
        ("LOS", 'B_.*',),
        ("LogSums", 'existing'),
    ]
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(15)
    m.maximize_loglike()

    m.calculate_parameter_covariance()
    m.parameter_summary()

    print(m.pfo()[['value','std_err','t_stat','robust_std_err','robust_t_stat']])
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import larch.numba as lx
import pytest
from pytest import approx

def test_dataset():
    d = lx.examples.MTC(format='dataset')

    assert d.CASEID == 'caseid'
    assert d.ALTID == 'altid'

    assert d.chose.CASEID == 'caseid'
    assert d.chose.ALTID == 'altid'

    assert d['chose'].CASEID == 'caseid'
    assert d['chose'].ALTID == 'altid'




__version__ = '5.6.1'

from .util.interface_info import Info, ipython_status
import sys
from .util.styles import css

info = Info('Larch', False, __version__)
version = Info('Larch', False, __version__, minimal=True)


def require_version(n):
    from . import __version__
    try:
        from packaging import version
    except:
        def int_from(x):
            import re
            nums = re.findall(r'\d+', x)
            if nums:
                return int(nums[0])
            return 0
        r = [int_from(i) for i in n.split(".")[:2]]
        v = [int_from(i) for i in __version__.split(".")[:2]]
        if v[0] > r[0]:
            return
        if v[0] < r[0]:
            raise ValueError("the installed larch is version {}".format(__version__))
        if len(r)>=2:
            if v[1] > r[1]:
                return
            if v[1] < r[1]:
                raise ValueError("the installed larch is version {}".format(__version__))
        if len(r)>=3:
            if v[2] > r[2]:
                return
            if v[2] < r[2]:
                raise ValueError("the installed larch is version {}".format(__version__))
    else:
        if version.parse(n) > version.parse(__version__):
            raise ValueError("the installed larch is version {}".format(__version__))


if 'IPython' in ipython_status():
    from .util.display import display
    try:
        # from .util.styles import stylesheet
        # stylesheet()
        # display(info)
        pass
    except:
        # print(repr(info))
        jupyter_active = False
    else:
        jupyter_active = True
else:
    jupyter_active = False
    # print(repr(info))


from .roles import P, X, PX
from .data_services import DataService
from .omx import OMX
from .data_services.dbf.dbf_reader import DBF

from .model import Model
from .dataframes import DataFrames
from .dataset import Dataset, DataTree

from .examples import example
from .util import figures
from .util.excel import ExcelWriter

_doctest_mode_ = False

_larch_self = sys.modules[__name__]

from xmle import Reporter, NumberedCaption
from xmle import load_metadata as read_metadata
from .workspace import make_reporter



def example(extract='m', estimate=False):
    import larch.numba as lx
    d = lx.examples.MTC(format='dataset')
    m = lx.Model(d)

    m.availability_var = 'avail'
    m.choice_ca_var = 'chose'

    from larch.roles import P, X

    m.utility_ca = (
        + X("totcost/hhinc") * P("costbyincome")
        + X("tottime * (altnum <= 4)") * P("motorized_time")
        + X("tottime * (altnum >= 5)") * P("nonmotorized_time")
        + X("ovtt/dist * (altnum <= 4)") * P("motorized_ovtbydist")
    )

    for a in [4,5,6]:
        m.utility_co[a] += X("hhinc") * P("hhinc#{}".format(a))

    for i in d['alt_names'][1:3]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            + X("vehbywrk") * P("vehbywrk_SR")
            + X("wkccbd+wknccbd") * P("wkcbd_"+name)
            + X("wkempden") * P("wkempden_"+name)
            + P("ASC_"+name)
        )

    for i in d['alt_names'][3:]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            + X("vehbywrk") * P("vehbywrk_"+name)
            + X("wkccbd+wknccbd") * P("wkcbd_"+name)
            + X("wkempden") * P("wkempden_"+name)
            + P("ASC_"+name)
        )

    m.ordering = (
        ('LOS', ".*cost.*", ".*time.*", ".*dist.*",),
        ('Zonal', "wkcbd.*", "wkempden.*",),
        ('Household', "hhinc.*", "vehbywrk.*",),
        ('ASCs', "ASC.*",),
    )
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.maximize_loglike()

    m.calculate_parameter_covariance()
    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import larch.numba as lx
    d = lx.examples.MTC(format='dataset')

    d1 = d.query_cases("numveh <= 1")
    d2 = d.query_cases("numveh > 1")
    m1 = lx.Model(d1, title="Cars<=1")
    m2 = lx.Model(d2, title="Cars>=2")

    from larch import P, X

    for m in (m1,m2):
        m.availability_var = 'avail'
        m.choice_ca_var = 'chose'
        m.utility_ca = (
            + X("totcost/hhinc") * P("costbyincome")
            + X("tottime * (altnum <= 4)") * P("motorized_time")
            + X("tottime * (altnum >= 5)") * P("nonmotorized_time")
            + X("ovtt/dist * (altnum <= 4)") * P("motorized_ovtbydist")
        )
        for a in [4,5,6]:
            m.utility_co[a] += X("hhinc") * P("hhinc#{}".format(a))
        for i in d['alt_names'][1:3]:
            name = str(i.values)
            a = int(i.altid)
            m.utility_co[a] += (
                + X("vehbywrk") * P("vehbywrk_SR")
                + X("wkccbd+wknccbd") * P("wkcbd_"+name)
                + X("wkempden") * P("wkempden_"+name)
                + P("ASC_"+name)
            )
        for i in d['alt_names'][3:]:
            name = str(i.values)
            a = int(i.altid)
            m.utility_co[a] += (
                + X("vehbywrk") * P("vehbywrk_"+name)
                + X("wkccbd+wknccbd") * P("wkcbd_"+name)
                + X("wkempden") * P("wkempden_"+name)
                + P("ASC_"+name)
            ) 
        m.ordering = (
            ('LOS', ".*cost.*", ".*time.*", ".*dist.*",),
            ('Zonal', "wkcbd.*", "wkempden.*",),
            ('Household', "hhinc.*", "vehbywrk.*",),
            ('ASCs', "ASC.*",),
        )
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    r1 = m1.maximize_loglike()
    r2 = m2.maximize_loglike()

    for m in (m1,m2):
        m.calculate_parameter_covariance()
        m.loglike_null()

    from larch.util.summary import joint_parameter_summary
    joint_parameter_summary([m1,m2])

    m2.utility_ca = m2.utility_ca.reformat_param('{}_2Cars')

    for a in m2.utility_co:
        m2.utility_co[a] = m2.utility_co[a].reformat_param(
            pattern='(ASC)', repl='ASC_2Cars'
        )

    mg = lx.ModelGroup([m1,m2])

    # TEST
    assert mg.loglike('null') == approx(-7309.600971749634)

    mg.set_cap()
    rg = mg.maximize_loglike()
    mg.calculate_parameter_covariance()

    rg

    # TEST
    assert rg.loglike == approx(-3406.7232213364787)
    assert rg.n_cases == 5029

    mg.ordering = (
        ('LOS', ".*cost.*", ".*time.*", ".*dist.*",),
        ('Zonal', "wkcbd.*", "wkempden.*",),
        ('Household', "hhinc.*", "vehbywrk.*",),
        ('ASCs', "ASC.*",),
    )
    mg.parameter_summary()

    # TEST
    import re
    mash = lambda x: re.sub('\\s+', ' ', str(x)).strip()
    assert mash(mg.parameter_summary().data) == mash('''
                                            Value    Std Err  t Stat Signif  Null Value
    Category  Parameter                                                                
    LOS       costbyincome                -0.0174     0.0117   -1.49                0.0
              costbyincome_2Cars           -0.102     0.0154   -6.60    ***         0.0
              motorized_time              -0.0196    0.00567   -3.46    ***         0.0
              motorized_time_2Cars        -0.0195    0.00496   -3.94    ***         0.0
              nonmotorized_time           -0.0442    0.00776   -5.69    ***         0.0
              nonmotorized_time_2Cars     -0.0471    0.00868   -5.43    ***         0.0
              motorized_ovtbydist          -0.100     0.0237   -4.23    ***         0.0
              motorized_ovtbydist_2Cars    -0.187     0.0309   -6.05    ***         0.0
    Zonal     wkcbd_Bike                    0.460      0.364    1.27                0.0
              wkcbd_SR2                     0.233      0.124    1.88                0.0
              wkcbd_SR3+                     1.04      0.192    5.42    ***         0.0
              wkcbd_Transit                  1.27      0.168    7.53    ***         0.0
              wkcbd_Walk                    0.100      0.250    0.40                0.0
              wkempden_Bike               0.00137    0.00124    1.11                0.0
              wkempden_SR2                0.00132   0.000392    3.36    ***         0.0
              wkempden_SR3+               0.00184   0.000461    3.99    ***         0.0
              wkempden_Transit            0.00274   0.000358    7.64    ***         0.0
              wkempden_Walk               0.00248   0.000730    3.39    ***         0.0
    Household hhinc#4                    -0.00101    0.00208   -0.48                0.0
              hhinc#5                    -0.00366    0.00527   -0.70                0.0
              hhinc#6                    -0.00232    0.00331   -0.70                0.0
              vehbywrk_Bike                -0.255      0.290   -0.88                0.0
              vehbywrk_SR                  -0.314     0.0740   -4.24    ***         0.0
              vehbywrk_Transit             -0.638      0.132   -4.82    ***         0.0
              vehbywrk_Walk                -0.395      0.198   -1.99      *         0.0
    ASCs      ASC_2Cars_Bike                -2.94      0.629   -4.67    ***         0.0
              ASC_2Cars_SR2                 -1.91      0.127  -15.06    ***         0.0
              ASC_2Cars_SR3+                -3.53      0.173  -20.43    ***         0.0
              ASC_2Cars_Transit             -1.47      0.320   -4.60    ***         0.0
              ASC_2Cars_Walk                -1.00      0.517   -1.94                0.0
              ASC_Bike                      -1.39      0.439   -3.17     **         0.0
              ASC_SR2                       -1.58      0.126  -12.57    ***         0.0
              ASC_SR3+                      -3.26      0.206  -15.84    ***         0.0
              ASC_Transit                  -0.747      0.271   -2.76     **         0.0
              ASC_Walk                      0.314      0.416    0.75                0.0
    ''')

    mg.estimation_statistics()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from .model import NumbaModel as Model
from .. import DataFrames, P, X, PX, OMX, DBF, Reporter, NumberedCaption, read_metadata, examples, util, __version__
from ..examples import example as _example
from ..data_warehouse import example_file
from ..model.model_group import ModelGroup

try:
    from ..dataset import Dataset, DataArray, DataTree, merge
except RuntimeError:
    pass

def example(*args, **kwargs):
    import importlib
    kwargs['larch'] = importlib.import_module(__name__)
    return _example(*args, **kwargs)



def example(extract='m', estimate=False):
    import larch.numba as lx
    m = lx.example(17)

    shared = m.graph.new_node(parameter='mu_shared', children=[2,3], name='Shared')
    motorized = m.graph.new_node(parameter='mu_moto', children=[1,shared,4], name='Motorized')
    nonmotorized= m.graph.new_node(parameter='mu_nonmoto', children=[5,6], name='Nonmotorized')

    m.ordering = (
        ("CostbyInc","costbyincome",),
        ("TravelTime",".*time.*",".*dist.*", ),
        ("Household","hhinc.*","vehbywrk.*",),
        ("Zonal","wkcbd.*","wkempden.*",),
        ("ASCs","ASC.*",),
    )
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.maximize_loglike()

    m.calculate_parameter_covariance()
    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import os
    import numpy as np
    import pandas as pd 
    import larch.numba as lx
    from larch import P, X

    hh, pp, tour, skims = lx.example(200, ['hh', 'pp', 'tour', 'skims'])

    from addicty import Dict

    Mode = Dict(
        DA = 1,
        SR = 2,
        Walk = 3,
        Bike = 4,
        Transit = 5,
    ).freeze()

    tour_dataset = lx.Dataset.from_idco(tour.set_index('TOURID'), alts=Mode)
    od_skims = lx.Dataset.from_omx(skims)

    dt = lx.DataTree(
        tour=tour_dataset,
        hh=hh.set_index('HHID'),
        person=pp.set_index('PERSONID'),
        od=od_skims,
        do=od_skims,
        relationships=(
            "tours.HHID @ hh.HHID",
            "tours.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ od.otaz",
            "tours.DTAZ @ od.dtaz",
            "hh.HOMETAZ @ do.dtaz",
            "tours.DTAZ @ do.otaz",
        ),
    )

    dt_work = dt.query_cases("TOURPURP == 1")

    m = lx.Model(datatree = dt_work)
    m.title = "Exampville Work Tour Mode Choice v1"

    m.utility_co[Mode.DA] = (
            + P.InVehTime * X.AUTO_TIME
            + P.Cost * X.AUTO_COST # dollars per mile
    )

    m.utility_co[Mode.SR] = (
            + P.ASC_SR
            + P.InVehTime * X.AUTO_TIME
            + P.Cost * (X.AUTO_COST * 0.5) # dollars per mile, half share
            + P("LogIncome:SR") * X("log(INCOME)")
    )

    m.utility_co[Mode.Walk] = (
            + P.ASC_Walk
            + P.NonMotorTime * X.WALK_TIME
            + P("LogIncome:Walk") * X("log(INCOME)")
    )

    m.utility_co[Mode.Bike] = (
            + P.ASC_Bike
            + P.NonMotorTime * X.BIKE_TIME
            + P("LogIncome:Bike") * X("log(INCOME)")
    )

    m.utility_co[Mode.Transit] = (
            + P.ASC_Transit
            + P.InVehTime * X.TRANSIT_IVTT
            + P.OutVehTime * X.TRANSIT_OVTT
            + P.Cost * X.TRANSIT_FARE
            + P("LogIncome:Transit") * X('log(INCOME)')
    )

    Car = m.graph.new_node(parameter='Mu:Car', children=[Mode.DA, Mode.SR], name='Car')
    NonMotor = m.graph.new_node(parameter='Mu:NonMotor', children=[Mode.Walk, Mode.Bike], name='NonMotor')
    Motor = m.graph.new_node(parameter='Mu:Motor', children=[Car, Mode.Transit], name='Motor')

    m.graph

    m.choice_co_code = 'TOURMODE'

    m.availability_co_vars = {
        Mode.DA: 'AGE >= 16',
        Mode.SR: 1,
        Mode.Walk: 'WALK_TIME < 60',
        Mode.Bike: 'BIKE_TIME < 60',
        Mode.Transit: 'TRANSIT_FARE>0',
    }

    m.choice_avail_summary()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(20) # improves optimization stability
    result = m.maximize_loglike()

    m.calculate_parameter_covariance()

    m.parameter_summary()

    m.estimation_statistics()

    report = lx.Reporter(title=m.title)

    report.append('# Parameter Summary')
    report.append(m.parameter_summary())
    report

    report << "# Estimation Statistics" << m.estimation_statistics()

    report << "# Utility Functions" << m.utility_functions()

    report.save(
        '/tmp/exampville_mode_choice.html', 
        overwrite=True, 
        metadata=m,
    )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



def example(extract='m', estimate=False):
    import pandas as pd
    import larch.numba as lx

    raw_data = pd.read_csv(lx.example_file('swissmetro.csv.gz')).rename_axis(index='CASEID')
    data = lx.Dataset.from_idco(raw_data, alts={1:'Train', 2:'SM', 3:'Car'})
    data

    m = lx.Model(data.query_cases("PURPOSE in (1,3) and CHOICE != 0"))

    m.title = "swissmetro example 02 (weighted logit)"

    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }
    m.choice_co_code = 'CHOICE'

    m.weight_co_var = "1.0*(GROUP==2)+1.2*(GROUP==3)"

    from larch.roles import P,X
    m.utility_co[1] = P("ASC_TRAIN")
    m.utility_co[2] = 0
    m.utility_co[3] = P("ASC_CAR")
    m.utility_co[1] += X("TRAIN_TT") * P("B_TIME")
    m.utility_co[2] += X("SM_TT") * P("B_TIME")
    m.utility_co[3] += X("CAR_TT") * P("B_TIME")
    m.utility_co[1] += X("TRAIN_CO*(GA==0)") * P("B_COST")
    m.utility_co[2] += X("SM_CO*(GA==0)") * P("B_COST")
    m.utility_co[3] += X("CAR_CO") * P("B_COST")

    m.ordering = [
        ("ASCs", 'ASC.*',),
        ("LOS", 'B_.*',),
    ]
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(15)
    m.maximize_loglike(method='SLSQP')

    m.calculate_parameter_covariance()
    m.parameter_summary()

    # TEST
    assert m.parameter_summary().data.to_markdown() == '''
    |                       |   Value |   Std Err |   t Stat | Signif   |   Null Value |
    |:----------------------|--------:|----------:|---------:|:---------|-------------:|
    | ('ASCs', 'ASC_CAR')   | -0.114  |  0.0407   |    -2.81 | **       |            0 |
    | ('ASCs', 'ASC_TRAIN') | -0.757  |  0.0528   |   -14.32 | ***      |            0 |
    | ('LOS', 'B_COST')     | -0.0112 |  0.00049  |   -22.83 | ***      |            0 |
    | ('LOS', 'B_TIME')     | -0.0132 |  0.000537 |   -24.62 | ***      |            0 |
    '''[1:-1]
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import os
import tables as tb
import pandas as pd

def MTC(format='dataframes'):
	from larch.dataframes import DataFrames
	from larch.data_warehouse import example_file
	ca = pd.read_csv(example_file('MTCwork.csv.gz'), index_col=('casenum', 'altnum'))
	ca['altnum'] = ca.index.get_level_values('altnum')
	dt = DataFrames(
		ca,
		ch="chose",
		crack=True,
		alt_codes=[1, 2, 3, 4, 5, 6],
		alt_names=['DA', 'SR2', 'SR3', 'TRANSIT', 'BIKE', 'WALK']
	)
	if format in ('dataset', 'datapool'):
		from ..dataset import Dataset, DataArray
		dataset = Dataset.from_dataframe(dt.data_co.rename_axis(index='caseid'))
		dataset = dataset.merge(Dataset.from_dataframe(dt.data_ce_as_ca().rename_axis(index=('caseid', 'altid'))))
		dataset['avail'] = DataArray(dt.data_av.values, dims=['caseid', 'altid'], coords=dataset.coords)
		dataset.coords['alt_names'] = DataArray(
			['DA', 'SR2', 'SR3+', 'Transit', 'Bike', 'Walk'],
			dims=['altid'],
		)
		dataset.CASEID = 'caseid'
		dataset.ALTID = 'altid'
		return dataset
	elif format == 'dataframes':
		dt.data_ce_as_ca("_avail_")
		return dt
	else:
		raise ValueError(f"undefined format {format}")


def EXAMPVILLE(format='dataframes', model='mode', cache_dir=None):
	if format == 'datatree' and model == 'mode':
		from ..examples import example
		from ..dataset import Dataset, DataArray, DataTree
		_hh, _pp, _tour, _skims = example(200, ['hh', 'pp', 'tour', 'skims'])
		tours = Dataset(
			_tour.set_index('TOURID'), caseid='TOURID',
		)
		tours.coords['altid'] = DataArray(
			[1, 2, 3, 4, 5], dims="_altid_",
		)
		tours.coords['altname'] = DataArray(
			['DA', 'SR', 'Walk', 'Bike', 'Transit'], dims="_altid_",
		)
		od_skims = Dataset.from_omx(_skims)
		hh = Dataset(_hh.set_index('HHID'))
		pp = Dataset(_pp.set_index('PERSONID'))
		tree = DataTree(
			tours=tours,
			hh=hh,
			pp=pp,
			od=od_skims,
			do=od_skims,
			root_node_name='tours',
			relationships=(
				"tours.HHID @ hh.HHID",
				"tours.PERSONID @ pp.PERSONID",
				"hh.HOMETAZ @ od.otaz",
				"tours.DTAZ @ od.dtaz",
				"hh.HOMETAZ @ do.dtaz",
				"tours.DTAZ @ do.otaz",
			),
		)
		return tree
	elif format == 'dataframes' and model == 'mode':
		from ..examples import example
		hh, pp, tour, skims = example(200, ['hh', 'pp', 'tour', 'skims'])
		raw = tour.merge(hh, on='HHID').merge(pp, on=('HHID', 'PERSONID'))
		raw["HOMETAZi"] = raw["HOMETAZ"] - 1
		raw["DTAZi"] = raw["DTAZ"] - 1
		raw = raw.join(skims.get_rc_dataframe(raw.HOMETAZi, raw.DTAZi))
		from ..dataframes import DataFrames
		return DataFrames(
			co=raw,
			alt_codes=[1, 2, 3, 4, 5],
			alt_names=['DA', 'SR', 'Walk', 'Bike', 'Transit'],
			ch_name='TOURMODE',
		)
	else:
		raise ValueError(f"undefined format {format} for Exampville {model} choice model")

def SWISSMETRO():
	from ..util.temporaryfile import TemporaryGzipInflation
	warehouse_dir = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', )
	from .service import DataService
	from .h5 import H5PodCO, H5PodCS
	warehouse_file = TemporaryGzipInflation(os.path.join(warehouse_dir, "swissmetro.h5.gz"))
	f = tb.open_file(warehouse_file, mode='r')
	idco = H5PodCO(f.root.larch.idco)
	stack = H5PodCS(
		[idco], ident='stack_by_mode', alts=[1,2,3],
		traveltime={1: "TRAIN_TT", 2: "SM_TT", 3: "CAR_TT"},
		cost={1: "TRAIN_CO*(GA==0)", 2: "SM_CO*(GA==0)", 3: "CAR_CO"},
		avail={1:'TRAIN_AV*(SP!=0)', 2:'SM_AV', 3:'CAR_AV*(SP!=0)'},
		choice={1: "CHOICE==1", 2: "CHOICE==2", 3: "CHOICE==3"},
	)
	return DataService(pods=[idco, stack], altids=[1,2,3], altnames=['Train', 'SM', 'Car'])


def ITINERARY_RAW():
	warehouse_file = os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', 'itinerary_data.csv.gz')
	import pandas
	return pandas.read_csv(warehouse_file)


def example_file(filename):
	warehouse_file = os.path.normpath( os.path.join( os.path.dirname(__file__), '..', 'data_warehouse', filename) )
	if os.path.exists(warehouse_file):
		return warehouse_file
	raise FileNotFoundError(f"there is no example data file '{warehouse_file}' in data_warehouse")



import numpy as np
import pandas as pd
import xarray as xr
import logging
from typing import NamedTuple
import numba as nb

from ..dim_names import CASEID as _CASEID, ALTID as _ALTID, CASEALT as _CASEALT, ALTIDX as _ALTIDX, CASEPTR as _CASEPTR

class _case_slice:

    def __get__(self, obj, objtype=None):
        self.parent = obj
        return self

    def __getitem__(self, idx):
        kwds = {}
        for k in self.parent._fields:
            attribute = getattr(self.parent, k)
            if attribute is not None and (len(k) == 2 or k == 'ce_caseptr'):
                kwds[k] = attribute[idx]
            else:
                kwds[k] = attribute
        return type(self.parent)(**kwds)


class DataArrays(NamedTuple):
    ch: np.ndarray
    av: np.ndarray
    wt: np.ndarray
    co: np.ndarray
    ca: np.ndarray
    ce_data: np.ndarray
    ce_altidx: np.ndarray
    ce_caseptr: np.ndarray

    cs = _case_slice()

    # @property
    # def alternatives(self):
    #     if self.alt_codes is not None:
    #         if self.alt_names is not None:
    #             return dict(zip(self.alt_codes, self.alt_names))
    #         else:
    #             return {i:str(i) for i in self.alt_codes}
    #     else:
    #         raise ValueError("alt_codes not defined")


def to_dataset(dataframes):
    caseindex_name = _CASEID
    altindex_name = _ALTID
    from ..dataset import Dataset
    from xarray import DataArray
    coords = {
        caseindex_name: dataframes.caseindex.values,
        altindex_name: dataframes.alternative_codes(),
    }
    ds = Dataset(coords=coords)
    if dataframes.data_co is not None:
        caseindex_name = dataframes.data_co.index.name
        ds.update(Dataset.from_dataframe(dataframes.data_co))
    if dataframes.data_ca is not None:
        caseindex_name = dataframes.data_ca.index.names[0]
        altindex_name = dataframes.data_ca.index.names[1]
        ds.update(Dataset.from_dataframe(dataframes.data_ca))
    alt_names = dataframes.alternative_names()
    if alt_names:
        ds.coords['alt_names'] = DataArray(alt_names, dims=(altindex_name,))
    return ds


def prepare_data(
        datasource,
        request,
        float_dtype=None,
        cache_dir=None,
        flows=None,
):
    """
    Load data from a DataTree into a computationally-formatted Dataset.

    Parameters
    ----------
    datasource : DataTree or Dataset
    request : Mapping or Model
    float_dtype : dtype, optional
    cache_dir : Path-like, optional
        Directory to cache sharrow flows.
    flows : dict, optional
        Collection of previously prepared flows.

    Returns
    -------
    model_dataset : Dataset
        Ready for computation.
    flows : dict
    """
    log = logging.getLogger("Larch")
    from ..dataset import Dataset, DataArray, DataTree
    if float_dtype is None:
        float_dtype = np.float64
    log.debug(f"building dataset from datashare coords: {datasource.coords}")
    model_dataset = Dataset(
        coords=datasource.coords,
    )
    try:
        if 'idcoVars' in datasource.subspaces:
            model_dataset.coords.update(datasource.subspaces['idcoVars'].coords)
    except AttributeError:
        pass
    model_dataset.CASEID = datasource.CASEID
    model_dataset.ALTID = datasource.ALTID

    from .model import NumbaModel # avoid circular import
    if isinstance(request, NumbaModel):
        alts = request.graph.elemental_names()
        if model_dataset.ALTID not in model_dataset.coords:
            model_dataset.coords[model_dataset.ALTID] = list(alts.keys())
        if 'alt_names' not in model_dataset.coords:
            model_dataset.coords['alt_names'] = DataArray(list(alts.values()), dims=(model_dataset.ALTID,))
        request = request.required_data()

    if flows is None:
        flows = {}

    if isinstance(datasource, (DataTree,)):
        log.debug(f"adopting existing DataTree")
        if not datasource.relationships_are_digitized:
            datasource.digitize_relationships(inplace=True)
        datatree = datasource
        datatree_co = datatree.idco_subtree()
    elif isinstance(datasource, Dataset):
        datatree = datasource.as_tree()
        if not datatree.relationships_are_digitized:
            datatree.digitize_relationships(inplace=True)
        datatree_co = datatree.idco_subtree()
    else:
        log.debug(f"initializing new DataTree")
        datatree = DataTree(main=datasource)
        datatree.digitize_relationships(inplace=True)
        datatree_co = datatree.idco_subtree()

    if 'co' in request:
        log.debug(f"requested co data: {request['co']}")
        model_dataset, flows['co'] = _prep_co(
            model_dataset,
            datatree_co,
            request['co'],
            tag='co',
            dtype=float_dtype,
            cache_dir=cache_dir,
            flow=flows.get('co'),
        )
    if 'ca' in request:
        log.debug(f"requested ca data: {request['ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            model_dataset, flows['ca'] = _prep_ca(
                model_dataset,
                datatree,
                request['ca'],
                tag='ca',
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get('ca'),
            )
        else:
            model_dataset, flows['ce'] = _prep_ce(
                model_dataset,
                datatree,
                request['ca'],
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get('ce'),
            )
    if 'choice_ca' in request:
        log.debug(f"requested choice_ca data: {request['choice_ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            log.debug(f"  loading choice_ca data from idca")
            model_dataset, flows['choice_ca'] = _prep_ca(
                model_dataset,
                datatree,
                request['choice_ca'],
                tag='ch',
                preserve_vars=False,
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get('choice_ca'),
            )
        else:
            log.debug(f"  loading choice_ca data from idce")
            model_dataset, flows['choice_ce'] = _prep_ce(
                model_dataset,
                datatree,
                request['choice_ca'],
                v_tag='choice_ca',
                s_tag='choice_ce',
                preserve_vars=False,
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get('choice_ce'),
                attach_indexes=False,
            )
            da_ch = DataArray(
                ce_to_dense(
                    model_dataset['choice_ce_data'].values,
                    model_dataset[model_dataset.ALTIDX].values,
                    model_dataset[model_dataset.CASEPTR].values,
                    datatree.n_alts
                ),
                dims=[datatree.CASEID, datatree.ALTID],
                coords={
                    datatree.CASEID: model_dataset.coords[datatree.CASEID],
                    datatree.ALTID: model_dataset.coords[datatree.ALTID],
                },
                name='ch',
            )
            model_dataset = model_dataset.drop_vars(['choice_ce_data'])
            model_dataset['ch'] = da_ch
    if 'choice_co_code' in request:
        log.debug(f"requested choice_co_code data: {request['choice_co_code']}")
        choicecodes = datasource[request['choice_co_code']]
        da_ch = DataArray(
            float_dtype(0),
            dims=[datatree.CASEID, datatree.ALTID],
            coords={
                datatree.CASEID: model_dataset.coords[datatree.CASEID],
                datatree.ALTID: model_dataset.coords[datatree.ALTID],
            },
            name='ch',
        )
        for i,a in enumerate(model_dataset.coords[datatree.ALTID]):
            da_ch[:, i] = (choicecodes == a)
        model_dataset = model_dataset.merge(da_ch)
    if 'choice_co' in request:
        log.debug(f"requested choice_co_vars data: {request['choice_co']}")
        da_ch = DataArray(
            float_dtype(0),
            dims=[datatree.CASEID, datatree.ALTID],
            coords={
                datatree.CASEID: model_dataset.coords[datatree.CASEID],
                datatree.ALTID: model_dataset.coords[datatree.ALTID],
            },
            name='ch',
        )
        for i,a in enumerate(model_dataset.alts_mapping):
            choice_expr = request['choice_co'][a]
            da_ch[:, i] = datatree_co.get_expr(choice_expr).values
        model_dataset = model_dataset.merge(da_ch)
    if 'choice_any' in request:
        log.debug(f"requested choice_any data: {request['choice_any']}")
        raise NotImplementedError('choice_any')

    if 'weight_co' in request:
        log.debug(f"requested weight_co data: {request['weight_co']}")
        model_dataset, flows['weight_co'] = _prep_co(
            model_dataset,
            datatree_co,
            [request['weight_co']],
            tag='wt',
            preserve_vars=False,
            dtype=float_dtype,
            cache_dir=cache_dir,
            flow=flows.get('weight_co'),
        )

    if 'avail_ca' in request:
        log.debug(f"requested avail_ca data: {request['avail_ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            model_dataset, flows['avail_ca'] = _prep_ca(
                model_dataset,
                datatree,
                request['avail_ca'],
                tag='av',
                preserve_vars=False,
                dtype=np.int8,
                cache_dir=cache_dir,
                flow=flows.get('avail_ca'),
            )
        else:
            if request['avail_ca'] in {'1', 'True', '1.0'} and model_dataset.CASEPTR is not None and model_dataset.ALTIDX is not None:
                da_av = DataArray(
                    ce_bool_to_dense(
                        model_dataset[model_dataset.ALTIDX].values,
                        model_dataset[model_dataset.CASEPTR].values,
                        datatree.n_alts,
                    ),
                    dims=[datatree.CASEID, datatree.ALTID],
                    coords={
                        datatree.CASEID: model_dataset.coords[datatree.CASEID],
                        datatree.ALTID: model_dataset.coords[datatree.ALTID],
                    },
                    name='av',
                )
            else:
                model_dataset, flows['avail_ce'] = _prep_ce(
                    model_dataset,
                    datatree,
                    request['avail_ca'],
                    v_tag='avail_ca',
                    s_tag='avail_ce',
                    preserve_vars=False,
                    dtype=np.int8,
                    cache_dir=cache_dir,
                    flow=flows.get('avail_ce'),
                    attach_indexes=False,
                )
                da_av = DataArray(
                    ce_to_dense(
                        model_dataset['avail_ce_data'].values,
                        model_dataset[model_dataset.ALTIDX].values,
                        model_dataset[model_dataset.CASEPTR].values,
                        datatree.n_alts,
                    ),
                    dims=[datatree.CASEID, datatree.ALTID],
                    coords={
                        datatree.CASEID: model_dataset.coords[datatree.CASEID],
                        datatree.ALTID: model_dataset.coords[datatree.ALTID],
                    },
                    name='av',
                )
                model_dataset = model_dataset.drop_vars(['avail_ce_data'])
            model_dataset['av'] = da_av
    if 'avail_co' in request:
        log.debug(f"requested avail_co data: {request['avail_co']}")
        av_co_expressions = {
            a: request['avail_co'].get(a, '0')
            for a in model_dataset.coords[datatree.ALTID].values
        }
        model_dataset, flows['avail_co'] = _prep_co(
            model_dataset,
            datatree_co,
            av_co_expressions,
            tag='av',
            preserve_vars=False,
            dtype=np.int8,
            dim_name=datatree.ALTID,
            cache_dir=cache_dir,
            flow=flows.get('avail_co'),
        )
    if 'avail_any' in request:
        log.debug(f"requested avail_any data: {request['avail_any']}")
        raise NotImplementedError('avail_any')

    return model_dataset, flows

def flownamer(tag, definition_spec, extra_hash_features=()):
    import hashlib, base64
    defs_hash = hashlib.md5()
    defs_hash.update(str(tag).encode("utf8"))
    for k, v in definition_spec.items():
        defs_hash.update(str(k).encode("utf8"))
        defs_hash.update(str(v).encode("utf8"))
    for k in extra_hash_features:
        defs_hash.update(str(k).encode("utf8"))
    return "pipeline_"+(base64.b32encode(defs_hash.digest())).decode().replace("=","")


def _prep_ca(
        model_dataset,
        shared_data_ca,
        vars_ca,
        tag='ca',
        preserve_vars=True,
        dtype=None,
        cache_dir=None,
        flow=None,
):
    from ..dataset import Dataset, DataArray, DataTree
    assert isinstance(shared_data_ca, DataTree)
    if isinstance(vars_ca, str):
        if not preserve_vars and vars_ca in shared_data_ca.root_dataset:
            proposal = shared_data_ca.root_dataset[vars_ca]
            if shared_data_ca.CASEID in proposal.dims and shared_data_ca.ALTID in proposal.dims:
                proposal = proposal.drop(list(proposal.coords)).rename(tag)
                return model_dataset.merge(proposal), flow
    if isinstance(vars_ca, str):
        vars_ca = {vars_ca: vars_ca}
    if not isinstance(vars_ca, dict):
        vars_ca = {i:i for i in vars_ca}
    flowname = flownamer(tag, vars_ca, shared_data_ca._hash_features())
    if flow is None or flowname != flow.name:
        flow = shared_data_ca.setup_flow(vars_ca, cache_dir=cache_dir, name=flowname)
    else:
        logging
    arr = flow.load(
        shared_data_ca,
        dtype=dtype,
    )
    caseid_dim = shared_data_ca.CASEID
    altid_dim = shared_data_ca.ALTID
    if preserve_vars or len(vars_ca)>1:
        arr = arr.reshape(
            model_dataset.dims.get(caseid_dim),
            model_dataset.dims.get(altid_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, altid_dim, f"var_{tag}"],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                altid_dim: model_dataset.coords[altid_dim],
                f"var_{tag}": list(vars_ca.keys()),
            },
            name=tag,
        )
    else:
        arr = arr.reshape(
            model_dataset.dims.get(caseid_dim),
            model_dataset.dims.get(altid_dim),
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, altid_dim],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                altid_dim: model_dataset.coords[altid_dim],
            },
            name=tag,
        )
    return model_dataset.merge(da), flow

def _prep_ce(
        model_dataset,
        datatree,
        vars_ca,
        v_tag='ca',
        s_tag='ce',
        preserve_vars=True,
        dtype=None,
        cache_dir=None,
        flow=None,
        attach_indexes=True,
):
    from ..dataset import Dataset, DataArray, DataTree
    assert isinstance(datatree, DataTree)
    if isinstance(vars_ca, str):
        if not preserve_vars and vars_ca in datatree.root_dataset:
            proposal = datatree.root_dataset[vars_ca]
            if datatree.CASEALT in proposal.dims:
                proposal = proposal.drop(list(proposal.coords)).rename(f"{s_tag}_data")
                return model_dataset.merge(proposal), flow
    if isinstance(vars_ca, str):
        vars_ca = {vars_ca: vars_ca}
    if not isinstance(vars_ca, dict):
        vars_ca = {i:i for i in vars_ca}
    flowname = flownamer(s_tag, vars_ca, datatree._hash_features())
    if flow is None or flowname != flow.name:
        flow = datatree.setup_flow(vars_ca, cache_dir=cache_dir, name=flowname)
    arr = flow.load(
        datatree,
        dtype=dtype,
    )
    casealt_dim = datatree.CASEALT
    if preserve_vars or len(vars_ca)>1:
        arr = arr.reshape(
            model_dataset.dims.get(casealt_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[casealt_dim, f"var_{v_tag}"],
            coords={
                casealt_dim: model_dataset.coords[casealt_dim],
                f"var_{v_tag}": list(vars_ca.keys()),
            },
            name=f"{s_tag}_data",
        )
    else:
        arr = arr.reshape(-1)
        da = DataArray(
            arr,
            dims=[casealt_dim],
            coords={
                casealt_dim: model_dataset.coords[casealt_dim],
            },
            name=f"{s_tag}_data",
        )
    model_dataset = model_dataset.merge(da)

    if attach_indexes:
        altidx = datatree.root_dataset.coords[datatree.ALTIDX]
        altidx = altidx.drop_vars(list(altidx.coords))
        model_dataset[datatree.ALTIDX] = altidx
        model_dataset.ALTIDX = datatree.ALTIDX
        caseptr = datatree.root_dataset[datatree.CASEPTR]
        caseptr = caseptr.drop_vars(list(caseptr.coords))
        model_dataset[datatree.CASEPTR] = caseptr
        model_dataset.CASEPTR = datatree.CASEPTR
        model_dataset = model_dataset.assign_coords({
            datatree.CASEID: DataArray(
                datatree.caseids(),
                dims=(datatree.CASEID),
            ),
            datatree.ALTID: DataArray(
                datatree.altids(),
                dims=(datatree.ALTID),
            ),
        })
    model_dataset.CASEID = datatree.CASEID
    model_dataset.ALTID = datatree.ALTID
    return model_dataset, flow


def _prep_co(
        model_dataset,
        shared_data_co,
        vars_co,
        tag='co',
        preserve_vars=True,
        dtype=None,
        dim_name=None,
        cache_dir=None,
        flow=None,
):
    from ..dataset import DataArray, DataTree
    assert isinstance(shared_data_co, DataTree)
    if not isinstance(vars_co, dict):
        vars_co = {i: i for i in vars_co}
    flowname = flownamer(tag, vars_co, shared_data_co._hash_features())
    if flow is None or flowname != flow.name:
        flow = shared_data_co.setup_flow(vars_co, cache_dir=cache_dir, name=flowname)
    arr = flow.load(
        shared_data_co,
        dtype=dtype,
    )
    caseid_dim = shared_data_co.CASEID
    if preserve_vars or len(vars_co)>1:
        if dim_name is None:
            dim_name = f"var_{tag}"
        arr = arr.reshape(
            model_dataset.dims.get(caseid_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, dim_name],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                dim_name: list(vars_co.keys()),
            },
            name=tag,
        )
    else:
        arr = arr.reshape(
            model_dataset.dims.get(caseid_dim),
        )
        da = DataArray(
            arr,
            dims=[caseid_dim],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
            },
            name=tag,
        )
    return model_dataset.merge(da), flow


@nb.njit
def ce_to_dense(ce_data, ce_altidx, ce_caseptr, n_alts):
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:,-1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0], n_alts, *ce_data.shape[1:])
    out = np.zeros(shape, dtype=ce_data.dtype)
    c = 0
    for row in range(ce_data.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
        a = ce_altidx[row]
        out[c,a,...] = ce_data[row,...]
    return out

@nb.njit
def ce_bool_to_dense(ce_altidx, ce_caseptr, n_alts):
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:,-1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0], n_alts)
    out = np.zeros(shape, dtype=np.int8)
    c = 0
    for row in range(ce_altidx.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
        a = ce_altidx[row]
        out[c,a,...] = 1
    return out



def example(extract='m', estimate=False):
    import os
    import numpy as np
    import xarray as xr
    from addicty import Dict
    from larch import P, X
    import larch.numba as lx

    hh, pp, tour, skims = lx.example(200, ['hh', 'pp', 'tour', 'skims'])

    exampville_mode_choice_file = lx.example(201, output_file='/tmp/exampville_mode_choice.html')
    m = lx.read_metadata(exampville_mode_choice_file)

    Mode = Dict(
        DA = 1,
        SR = 2,
        Walk = 3,
        Bike = 4,
        Transit = 5,
    ).freeze()

    tour_dataset = lx.Dataset.from_idco(tour.set_index('TOURID'), alts=Mode)
    od_skims = lx.Dataset.from_omx(skims)

    dt = lx.DataTree(
        tour=tour_dataset.query_cases("TOURPURP == 1"),
        hh=hh.set_index('HHID'),
        person=pp.set_index('PERSONID'),
        od=od_skims,
        do=od_skims,
        relationships=(
            "tours.HHID @ hh.HHID",
            "tours.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ od.otaz",
            "tours.DTAZ @ od.dtaz",
            "hh.HOMETAZ @ do.dtaz",
            "tours.DTAZ @ do.otaz",
        ),
    )

    logsums = lx.DataArray.zeros(
        dt.caseids(),
        skims.TAZ_ID,
        name='logsums',
    )

    for i,dtaz in enumerate(logsums.TAZ_ID):
        m.datatree = dt.replace_datasets(
            tour=dt.root_dataset.assign(
                DTAZ=xr.full_like(dt['DTAZ'], dtaz)
            ),
        )
        logsums.loc[dict(TAZ_ID=dtaz)] = m.logsums()

    logsums.to_zarr('/tmp/logsums.zarr.zip')

    lx.DataArray.from_zarr('/tmp/logsums.zarr.zip')
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import os
import subprocess


def make_estimation_data(directory="test"):
	if os.path.exists(directory):
		check_file = os.path.join(directory, ".gitignore")
		if not os.path.exists(check_file):
			import importlib
			asim_create = importlib.import_module("activitysim.cli.create")
			asim_create.get_example("example_estimation_sf", directory)
			os.chdir(directory)
			cp = subprocess.run(
				[
					"activitysim",
					"run",
					"-c",
					"configs_estimation/configs",
					"-c",
					"configs",
					"-o",
					"output",
					"-d",
					"data_sf",
				],
				capture_output=True,
			)
			with open(check_file, 'rt') as f:
				f.write("**/*.csv\n")
				f.write("**/*.txt\n")
				f.write("**/*.yaml\n")
				f.write("**/.gitignore\n")
		else:
			print(f"using existing directory `{directory}`")
			os.chdir(directory)



def example(extract='m', estimate=False):
    import os, gzip
    import numpy as np, pandas as pd, xarray as xr
    import larch.numba as lx

    df = pd.read_csv(lx.example_file("MTCwork.csv.gz"), index_col=['casenum','altnum'])

    df.head(15)

    ds = lx.Dataset.from_idca(df)
    ds

    ds = ds.set_altnames({
        1:'DA', 2:'SR2', 3:'SR3+', 4:'Transit', 5:'Bike', 6:'Walk',
    })
    ds
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]




# Example code extracted automatically from Larch documentation.
# Do not edit these files manually.
from ._000_mtc_data import example as _000_mtc_data
from ._001_mnl import example as _001_mnl
from ._002_mtc import example as _002_mtc
from ._017A_mnl_segmented import example as _017A_mnl_segmented
from ._017_mnl_final import example as _017_mnl_final
from ._022_nl import example as _022_nl
from ._028_nl import example as _028_nl
from ._101_swissmetro_mnl import example as _101_swissmetro_mnl
from ._102_swissmetro_weighted import example as _102_swissmetro_weighted
from ._109_swissmetro_nl import example as _109_swissmetro_nl
from ._200_exampville import example as _200_exampville
from ._201_exville_mode_choice import example as _201_exville_mode_choice
from ._202_exville_mc_logsums import example as _202_exville_mc_logsums
from ._203_exville_dest import example as _203_exville_dest
from ._300_itinerary import example as _300_itinerary
from ._301_itin_mnl import example as _301_itin_mnl

ex = {}
ex[0] = _000_mtc_data
ex[1] = _001_mnl
ex[2] = _002_mtc
ex['017A'] = _017A_mnl_segmented
ex[17] = _017_mnl_final
ex[22] = _022_nl
ex[28] = _028_nl
ex[101] = _101_swissmetro_mnl
ex[102] = _102_swissmetro_weighted
ex[109] = _109_swissmetro_nl
ex[200] = _200_exampville
ex[201] = _201_exville_mode_choice
ex[202] = _202_exville_mc_logsums
ex[203] = _203_exville_dest
ex[300] = _300_itinerary
ex[301] = _301_itin_mnl



def example(extract='m', estimate=False):
    import larch.numba as lx

    d = lx.Dataset.from_idce(
        pd.read_csv(lx.example_file("arc"), index_col=['id_case','id_alt']),
    )

    m = lx.Model(datatree=d)

    v = [
        "timeperiod==2",
        "timeperiod==3",
        "timeperiod==4",
        "timeperiod==5",
        "timeperiod==6",
        "timeperiod==7",
        "timeperiod==8",
        "timeperiod==9",
        "carrier==2",
        "carrier==3",
        "carrier==4",
        "carrier==5",
        "equipment==2",
        "fare_hy",    
        "fare_ly",    
        "elapsed_time",  
        "nb_cnxs",       
    ]


    from larch.roles import PX
    m.utility_ca = sum(PX(i) for i in v)
    m.choice_ca_var = 'choice'
    m.availability_var = 1
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.maximize_loglike()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import numpy as np
import pandas as pd

from .abstract_model import AbstractChoiceModel
from . import persist_flags
from ..exceptions import ParameterNotInModelWarning
from .constraints import ParametricConstraintList
from collections.abc import MutableSequence

class ModelGroup(AbstractChoiceModel, MutableSequence):

	constraints = ParametricConstraintList()

	def __init__(
			self,
			models,
			*,
			parameters=None,
			frame=None,
			title=None,
			dataservice=None,
			constraints=None,
	):
		super().__init__(
			parameters=parameters,
			frame=frame,
			title=title,
		)
		self._k_models = list()
		for model in models:
			if isinstance(model, ModelGroup):
				self._k_models.extend(model._k_models)
			else:
				self._k_models.append(model)
		self._dataservice = dataservice
		self._dataframes = None
		self._mangled = True
		self.constraints = constraints

	def __getitem__(self, x):
		return self._k_models[x]

	def __setitem__(self, i, value):
		assert isinstance(value, AbstractChoiceModel)
		self._k_models[i] = value

	def __delitem__(self, x):
		del self._k_models[x]

	def __len__(self):
		return len(self._k_models)

	def __contains__(self, x):
		return (x in self.pf.index)

	def insert(self, i, value):
		assert isinstance(value, AbstractChoiceModel)
		self._k_models.insert(i,value)

	@property
	def dataframes(self):
		return 'access dataframes of group members individually'

	@property
	def n_cases(self):
		"""int : Total number of cases in the attached data of all grouped models."""
		return sum(k.n_cases for k in self._k_models)

	def total_weight(self):
		"""
		The total weight of cases in the attached data of all grouped models.

		Returns
		-------
		float
		"""
		return sum(k.total_weight() for k in self._k_models)

	def unmangle(self, force=False):
		super().unmangle(force)
		for k in self._k_models:
			rows_to_import = [i for i in k.pf.index if i not in self._frame.index]
			if len(rows_to_import):
				new_frame = pd.concat([
					self._frame, k.pf.loc[rows_to_import]
				]).astype(self._frame.dtypes)

				if not (
					np.array_equal(new_frame['value'].values, self._frame['value'].values)
					and np.array_equal(new_frame.index, self._frame.index)
				):
					frame_values_have_changed = True
				else:
					frame_values_have_changed = False

				self._frame = new_frame
				if frame_values_have_changed:
					self._frame_values_have_changed()

			k.unmangle(force)

	def set_values(self, values=None, *, respect_holdfast=True, **kwargs):
		"""
		Set the parameter values for one or more parameters.

		Parameters
		----------
		values : {'null', 'init', 'best', array-like, dict, scalar}, optional
			New values to set for the parameters.
			If 'null' or 'init', the current values are set
			equal to the null or initial values given in
			the 'nullvalue' or 'initvalue' column of the
			parameter frame, respectively.
			If 'best', the current values are set equal to
			the values given in the 'best' column of the
			parameter frame, if that columns exists,
			otherwise a ValueError exception is raised.
			If given as array-like, the array must be a
			vector with length equal to the length of the
			parameter frame, and the given vector will replace
			the current values.  If given as a dictionary,
			the dictionary is used to update `kwargs` before
			they are processed.
		kwargs : dict
			Any keyword arguments (or if `values` is a
			dictionary) are used to update the included named
			parameters only.  A warning will be given if any key of
			the dictionary is not found among the existing named
			parameters in the parameter frame, and the value
			associated with that key is ignored.  Any parameters
			not named by key in this dictionary are not changed.

		Notes
		-----
		Setting parameters both in the `values` argument and
		through keyword assignment is not explicitly disallowed,
		although it is not recommended and may be disallowed in the future.

		Raises
		------
		ValueError
			If setting to 'best' but there is no 'best' column in
			the `pf` parameters DataFrame.
		"""
		super().set_values(values=values, respect_holdfast=respect_holdfast, **kwargs)
		self._frame_values_have_changed()

	def _frame_values_have_changed(self):
		vals = self.pf.value.copy()
		import warnings
		with warnings.catch_warnings():
			warnings.simplefilter("ignore", category=ParameterNotInModelWarning)
			for k in self._k_models:
				k.set_values(**vals)
				k._frame = k._frame.copy()
		for k in self._k_models:
			k._frame_values_have_changed()

	def __prep_for_compute(self, x=None):
		self.unmangle()
		if x is not None:
			self.set_values(x)
		vals = self.pf.value.copy()
		import warnings
		with warnings.catch_warnings():
			warnings.simplefilter("ignore", category=ParameterNotInModelWarning)
			for k in self._k_models:
				k.set_values(**vals)

	def load_data(self, dataservice=None, autoscale_weights=True, log_warnings=True):
		for k in self._k_models:
			k.load_data(
				dataservice=dataservice,
				autoscale_weights=autoscale_weights,
				log_warnings=log_warnings,
			)

	def loglike(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=True,
			probability_only=False,
	):
		"""
		Compute a log likelihood value.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll' and the first derivative by key 'dll'.
			Other arrays are also included if `persist` is set to True.

		"""

		if start_case != 0:
			raise NotImplementedError('start_case != 0')
		if stop_case != -1:
			raise NotImplementedError('stop_case != -1')
		if step_case != 1:
			raise NotImplementedError('step_case != 1')
		if leave_out != -1:
			raise NotImplementedError('leave_out != -1')
		if keep_only != -1:
			raise NotImplementedError('keep_only != -1')
		if subsample != -1:
			raise NotImplementedError('subsample != -1')
		if probability_only:
			raise NotImplementedError('probability_only != False')

		from ..util import dictx
		self.__prep_for_compute(x)
		ll2_parts = [m.loglike(persist=persist) for m in self._k_models]
		if not persist:
			result = sum(ll2_parts)
			self._check_if_best(result)
			return result
		ll2 = dictx(
			ll=sum(y.ll for y in ll2_parts),
		)
		if persist & persist_flags.PERSIST_PARTS:
			ll2.ll_parts = [y.ll for y in ll2_parts]
		for key in ll2_parts[0].keys():
			if key != 'll':
				ll2[key] = list(y[key] for y in ll2_parts)
		self._check_if_best(ll2.ll)
		return ll2


	def loglike2(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=True,
			probability_only=False,

	):
		"""
		Compute a log likelihood value and its first derivative.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll' and the first derivative by key 'dll'.
			Other arrays are also included if `persist` is set to True.

		"""

		if start_case != 0:
			raise NotImplementedError('start_case != 0')
		if stop_case != -1:
			raise NotImplementedError('stop_case != -1')
		if step_case != 1:
			raise NotImplementedError('step_case != 1')
		if leave_out != -1:
			raise NotImplementedError('leave_out != -1')
		if keep_only != -1:
			raise NotImplementedError('keep_only != -1')
		if subsample != -1:
			raise NotImplementedError('subsample != -1')
		if probability_only:
			raise NotImplementedError('probability_only != False')

		from ..util import dictx
		self.__prep_for_compute(x)
		ll2_parts = [m.loglike2(persist=persist) for m in self._k_models]
		dll = ll2_parts[0].dll
		for y in ll2_parts[1:]:
			dll = dll.add(y.dll, fill_value=0)
		dll = dll[self.pf.index]
		ll2 = dictx(
			ll=sum(y.ll for y in ll2_parts),
			dll=dll,
		)
		for key in ll2_parts[0].keys():
			if key not in {'ll','dll'}:
				ll2[key] = list(y[key] for y in ll2_parts)
		self._check_if_best(ll2.ll)
		return ll2

	def doctor(
			self,
			repair_ch_av=None,
			repair_ch_zq=None,
			repair_asc=None,
			repair_noch_nowt=None,
			repair_nan_wt=None,
			repair_nan_data_co=None,
			verbose=3,
	):
		problems = []
		for k in self._k_models:
			k.unmangle(True)
			problems.append(k.doctor(
				repair_ch_av=repair_ch_av,
				repair_ch_zq=repair_ch_zq,
				repair_asc=repair_asc,
				repair_noch_nowt=repair_noch_nowt,
				repair_nan_wt=repair_nan_wt,
				repair_nan_data_co=repair_nan_data_co,
				verbose=verbose,
			))
		return problems

	def to_xlsx(self, filename, save_now=True, **kwargs):
		from ..util.excel import _make_excel_writer
		_make_excel_writer(self, filename, save_now=save_now, **kwargs)



def example(extract='m', estimate=False):
    import gzip, os, pandas as pd
    import larch.numba as lx

    itin = pd.read_csv(lx.example_file("arc"), index_col=['id_case','id_alt'])
    itin.info()

    d = lx.Dataset.from_idca(itin, crack=True)
    d
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]




class TouchNotify():
	"""
	A mixin to execute a notification callback any time a
	mutate-possible command is called.
	"""

	def __init__(self, *args, touch_callback_name='mangle', **kwargs):
		self._touch_name = touch_callback_name
		self._touch = lambda: None
		super().__init__(*args, **kwargs)

	def __set_name__(self, owner, name):
		# self : compiledmethod
		# owner : parent class that will have `self` as a member
		# name : the name of the attribute that `self` will be
		self.public_name = name
		self.private_name = '_touchable_' + name

	def __get__(self, obj, objtype=None):
		# self : TouchNotify
		# obj : instance of parent class that has `self` as a member, or None
		# objtype : class of `obj`
		result = getattr(obj, self.private_name, None)
		if result is not None:
			result.set_touch_callback(getattr(obj, self._touch_name, lambda: None))
		return result

	def __set__(self, obj, value):
		# self : TouchNotify
		# obj : instance of parent class that has `self` as a member
		# value : the new value that is trying to be assigned
		if not (isinstance(value, type(self)) or value is None):
			value = type(self)(value)
		setattr(obj, self.private_name, value)

	def __delete__(self, obj):
		# self : TouchNotify
		# obj : instance of parent class that has `self` as a member
		setattr(obj, self.private_name, None)

	def get_touch_callback(self):
		return self._touch

	def set_touch_callback(self, callback):
		if callable(callback):
			self._touch = callback
		elif callback is None:
			self._touch = lambda: None
		else:
			raise TypeError('callback must be callable')

	def del_touch_callback(self):
		self._touch = lambda: None

	def touch(self):
		self._touch()

	def __setitem__(self, key, value):
		ret = super().__setitem__(key, value)
		try:
			self.__touch()
		except AttributeError:
			pass # on init, may not yet have _touch
		return ret

	def __setattr__(self, key, value):
		ret = super().__setattr__(key, value)
		try:
			self.__touch()
		except AttributeError:
			pass # on init, may not yet have _touch
		return ret

	def __delitem__(self, key):
		ret = super().__delitem__(key)
		self._touch()
		return ret

	def __delattr__(self, key):
		ret = super().__delattr__(key)
		self._touch()
		return ret

	def setdefault(self, *args, **kwargs):
		ret = super().setdefault(*args, **kwargs)
		self._touch()
		return ret

	def append(self, *args, **kwargs):
		ret = super().append(*args, **kwargs)
		self._touch()
		return ret

	def extend(self, *args, **kwargs):
		ret = super().extend(*args, **kwargs)
		self._touch()
		return ret

	def update(self, *args, **kwargs):
		ret = super().update(*args, **kwargs)
		self._touch()
		return ret

	def insert(self, *args, **kwargs):
		ret = super().insert(*args, **kwargs)
		self._touch()
		return ret

	def remove(self, *args, **kwargs):
		ret = super().remove(*args, **kwargs)
		self._touch()
		return ret

	def pop(self, *args, **kwargs):
		ret = super().pop(*args, **kwargs)
		self._touch()
		return ret

	def popitem(self, *args, **kwargs):
		ret = super().popitem(*args, **kwargs)
		self._touch()
		return ret

	def clear(self, *args, **kwargs):
		ret = super().clear(*args, **kwargs)
		self._touch()
		return ret

	def __call__(self, *args, **kwargs):
		ret = super().__call__(*args, **kwargs)
		self._touch()
		return ret

	def __iadd__(self, other):
		ret = super().__iadd__(other)
		self._touch()
		return ret

	def __isub__(self, other):
		ret = super().__isub__(other)
		self._touch()
		return ret

	def __imul__(self, other):
		ret = super().__imul__(other)
		self._touch()
		return ret

	def __idiv__(self, other):
		ret = super().__idiv__(other)
		self._touch()
		return ret

	def __iand__(self, other):
		ret = super().__iand__(other)
		self._touch()
		return ret

	def __ior__(self, other):
		ret = super().__ior__(other)
		self._touch()
		return ret

	def __ixor__(self, other):
		ret = super().__ixor__(other)
		self._touch()
		return ret

	def __getstate__(self):
		return {}

	def __setstate__(self, state):
		self._touch = lambda: None



import logging
import warnings
import numpy as np
import pandas as pd
import xarray as xr
import numba as nb
import pathlib
from collections.abc import Mapping
from xarray.core import dtypes
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    DefaultDict,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

try:
    from sharrow import Dataset as _sharrow_Dataset
    from sharrow import DataArray as _sharrow_DataArray
    from sharrow import DataTree as _sharrow_DataTree
except ImportError:
    warnings.warn("larch.dataset requires the sharrow library")
    class _noclass:
        pass
    _sharrow_Dataset = xr.Dataset
    _sharrow_DataArray = xr.DataArray
    _sharrow_DataTree = _noclass

from .dim_names import CASEID as _CASEID, ALTID as _ALTID, CASEALT as _CASEALT, ALTIDX as _ALTIDX, CASEPTR as _CASEPTR

class DataArray(_sharrow_DataArray):

    __slots__ = ()

    def __init__(self, *args, caseid=None, altid=None, **kwargs):
        super().__init__(*args, **kwargs)
        if caseid is not None and caseid in self.dims:
            self.CASEID = caseid
        if altid is not None and altid in self.dims:
            self.ALTID = altid

    @classmethod
    def zeros(cls, *coords, dtype=np.float64, name=None, attrs=None):
        """
        Construct a dataset filled with zeros.

        Parameters
        ----------
        coords : Tuple[array-like]
            A sequence of coordinate vectors.  Ideally each should have a
            `name` attribute that names a dimension, otherwise placeholder
            names are used.
        dtype : dtype, default np.float64
            dtype of the new array. If omitted, it defaults to np.float64.
        name : str or None, optional
            Name of this array.
        attrs : dict_like or None, optional
            Attributes to assign to the new instance. By default, an empty
            attribute dictionary is initialized.

        Returns
        -------
        DataArray
        """
        dims = []
        shape = []
        coo = {}
        for n, c in enumerate(coords):
            i = getattr(c, "name", f"dim_{n}")
            dims.append(i)
            shape.append(len(c))
            coo[i] = c
        return cls(
            data=np.zeros(shape, dtype=dtype),
            coords=coo,
            dims=dims,
            name=name,
            attrs=attrs,
        )

    def to_zarr(self, store=None, *args, **kwargs):
        """
        Write DataArray contents to a zarr store.

        All parameters are passed directly to :py:meth:`xarray.Dataset.to_zarr`.

        Notes
        -----
        Only xarray.Dataset objects can be written to netCDF files, so
        the xarray.DataArray is converted to a xarray.Dataset object
        containing a single variable. If the DataArray has no name, or if the
        name is the same as a coordinate name, then it is given the name
        ``"__xarray_dataarray_variable__"``.

        See Also
        --------
        Dataset.to_zarr
        """
        from xarray.backends.api import DATAARRAY_VARIABLE

        if self.name is None:
            # If no name is set then use a generic xarray name
            dataset = self.to_dataset(name=DATAARRAY_VARIABLE)
        else:
            # No problems with the name - so we're fine!
            dataset = self.to_dataset()

        if isinstance(store, (str, pathlib.Path)) and len(args)==0:
            if str(store).endswith('.zip'):
                import zarr
                with zarr.ZipStore(store, mode='w') as zstore:
                    dataset.to_zarr(zstore, **kwargs)
                return

        return dataset.to_zarr(*args, **kwargs)

    @classmethod
    def from_zarr(cls, *args, name=None, **kwargs):
        dataset = xr.open_zarr(*args, **kwargs)
        if name is None:
            names = set(dataset.variables) - set(dataset.coords)
            if len(names) == 1:
                name = names.pop()
            else:
                raise ValueError("cannot infer name to load")
        return cls(dataset[name])

    @property
    def CASEID(self):
        result = self.attrs.get(_CASEID, None)
        if result is None:
            warnings.warn("no defined CASEID")
            return _CASEID
        return result

    @CASEID.setter
    def CASEID(self, dim_name):
        if dim_name not in self.dims:
            raise ValueError(f"{dim_name} not in dims")
        self.attrs[_CASEID] = dim_name

    @property
    def ALTID(self):
        result = self.attrs.get(_ALTID, None)
        if result is None:
            warnings.warn("no defined ALTID")
            return _ALTID
        return result

    @ALTID.setter
    def ALTID(self, dim_name):
        self.attrs[_ALTID] = dim_name

    @property
    def n_cases(self):
        try:
            i = self.dims.index(self.CASEID)
        except ValueError:
            logging.getLogger().error(f"missing {self.CASEID!r} among dims {self.dims}")
            raise
        return self.shape[i]

    @property
    def n_alts(self):
        if self.ALTID in self.dims:
            return self.shape[self.dims.index(self.ALTID)]
        if 'n_alts' in self.attrs:
            return self.attrs['n_alts']
        raise ValueError('no n_alts set')

    @property
    def alts_mapping(self):
        """Dict[int,str] : Mapping of alternative codes to names"""
        a = self.coords[self.ALTID]
        if 'alt_names' in a.coords:
            return dict(zip(a.values, a.coords['alt_names'].values))
        else:
            return dict(zip(a.values, a.values))

    def _repr_html_(self):
        html = super()._repr_html_()
        html = html.replace("sharrow.DataArray", "larch.DataArray")
        html = html.replace("xarray.DataArray", "larch.DataArray")
        return html

    def __repr__(self):
        r = super().__repr__()
        r = r.replace("sharrow.DataArray", "larch.DataArray")
        r = r.replace("xarray.DataArray", "larch.DataArray")
        return r

    def value_counts(self, index_name='index'):
        """
        Count the number of times each unique value appears in the array.

        Parameters
        ----------
        index_name : str, default 'index'
            Name of index dimension in result.

        Returns
        -------
        DataArray
        """
        values, freqs = np.unique(self, return_counts=True)
        return self.__class__(freqs, dims=index_name, coords={index_name:values})

    def clip(self, *args, **kwargs):
        """
        Return an array whose values are limited to ``[min, max]``.
        At least one of max or min must be given.

        Refer to `numpy.clip` for full documentation.

        See Also
        --------
        numpy.clip : equivalent function
        """
        out = super().clip(*args, **kwargs)
        out.__class__ = self.__class__
        return out


class Dataset(_sharrow_Dataset):
    """
    A xarray.Dataset extended interface for use with Larch.

    A Dataset consists of variables, coordinates and attributes which
    together form a self describing dataset.

    Dataset implements the mapping interface with keys given by variable
    names and values given by DataArray objects for each variable name.

    One dimensional variables with name equal to their dimension are
    index coordinates used for label based indexing.

    For Larch, one dimension of each Dataset must typiccally be named '_caseid_',
    and this dimension is used to identify the individual discrete choice
    observations or simulations in the data. The `caseid` argument can be
    used to set an existing dimension as '_caseid_' on Dataset construction.

    Parameters
    ----------
    data_vars : dict-like, optional
        A mapping from variable names to :py:class:`~xarray.DataArray`
        objects, :py:class:`~xarray.Variable` objects or to tuples of
        the form ``(dims, data[, attrs])`` which can be used as
        arguments to create a new ``Variable``. Each dimension must
        have the same length in all variables in which it appears.

        The following notations are accepted:

        - mapping {var name: DataArray}
        - mapping {var name: Variable}
        - mapping {var name: (dimension name, array-like)}
        - mapping {var name: (tuple of dimension names, array-like)}
        - mapping {dimension name: array-like}
          (it will be automatically moved to coords, see below)

        Each dimension must have the same length in all variables in
        which it appears.

    coords : dict-like, optional
        Another mapping in similar form as the `data_vars` argument,
        except the each item is saved on the dataset as a "coordinate".
        These variables have an associated meaning: they describe
        constant/fixed/independent quantities, unlike the
        varying/measured/dependent quantities that belong in
        `variables`. Coordinates values may be given by 1-dimensional
        arrays or scalars, in which case `dims` do not need to be
        supplied: 1D arrays will be assumed to give index values along
        the dimension with the same name.

        The following notations are accepted:

        - mapping {coord name: DataArray}
        - mapping {coord name: Variable}
        - mapping {coord name: (dimension name, array-like)}
        - mapping {coord name: (tuple of dimension names, array-like)}
        - mapping {dimension name: array-like}
          (the dimension name is implicitly set to be the same as the
          coord name)

        The last notation implies that the coord name is the same as
        the dimension name.

    attrs : dict-like, optional
        Global attributes to save on this dataset.

    caseid : str, optional, keyword only
        This named dimension will be marked as the '_caseid_' dimension.

    alts : str or Mapping or array-like, keyword only
        If given as a str, this named dimension will be marked as the
        '_altid_' dimension.  Otherwise, give a Mapping that defines
        alternative names and (integer) codes or an array of codes.
    """

    __slots__ = ('_flow_library')

    def __new__(cls, *args, caseid=None, alts=None, **kwargs):
        import logging
        logging.getLogger("sharrow").debug(f"NEW INSTANCE {cls}")
        obj = super().__new__(cls)
        super(cls, obj).__init__(*args, **kwargs)
        return cls.__initialize_for_larch(obj, caseid, alts)

    @classmethod
    def __initialize_for_larch(cls, obj, caseid=None, alts=None):
        """

        Parameters
        ----------
        obj : Dataset
            The dataaset being initialized.
        caseid : str, optional
            The name of a dimension referencing cases.
        alts : Mapping or str or array-like, optional
            If given as a mapping, links alternative codes to names.
            A string names a dimension that defines the alternatives.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        obj._flow_library = {}
        if caseid is not None:
            if caseid not in obj.dims:
                raise ValueError(f"no dim named '{caseid}' to make into {_CASEID}")
            obj.attrs[_CASEID] = caseid
        if isinstance(alts, pd.Series):
            alts_dim_name = alts.name or alts.index.name or '_altid_'
            alts_k = DataArray(
                alts.index, dims=alts_dim_name,
            )
            alts_v = DataArray(
                alts.values, dims=alts_dim_name,
            )
        elif isinstance(alts, Mapping):
            alts_dim_name = '_altid_'
            alts_k = DataArray(
                list(alts.keys()), dims=alts_dim_name,
            )
            alts_v = DataArray(
                list(alts.values()), dims=alts_dim_name,
            )
        elif isinstance(alts, str):
            alts_dim_name = alts
            alts_k = alts_v = None
        elif alts is None:
            alts_dim_name = None
            alts_k = alts_v = None
        else:
            alts_dim_name = getattr(alts, 'name', '_altid_')
            alts_v = np.asarray(alts).reshape(-1)
            alts_k = None
        if alts_dim_name:
            obj.attrs[_ALTID] = alts_dim_name
        if alts_k is not None:
            if np.issubdtype(alts_v, np.integer) and not np.issubdtype(alts_k, np.integer):
                obj.coords[alts_dim_name] = alts_v
                obj.coords['alt_names'] = alts_k
            else:
                obj.coords[alts_dim_name] = alts_k
                obj.coords['alt_names'] = alts_v
        elif alts_v is not None:
            obj.coords[alts_dim_name] = alts_v
        return obj

    @classmethod
    def construct(cls, source, caseid=None, alts=None):
        """
        A generic constructor for creating Datasets from various similar objects.

        Parameters
        ----------
        source : pandas.DataFrame, pyarrow.Table, xarray.Dataset, or Sequence[str]
            The source from which to create a Dataset.  DataFrames and Tables
            are converted to Datasets that have one dimension (the rows) and
            seperate variables for each of the columns.  A list of strings
            creates a dataset with those named empty variables.
        caseid : str, optional
            The name of a dimension referencing cases.
        alts : Mapping or str or array-like, optional
            If given as a mapping, links alternative codes to names.
            A string names a dimension that defines the alternatives.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        if isinstance(source, pd.DataFrame):
            source = cls.from_dataframe(source)
        else:
            source = super().construct(source)
        return cls.__initialize_for_larch(source, caseid, alts)

    def _set_sparse_data_from_dataframe_gcxs(
        self, idx: pd.Index, arrays: List[Tuple[Hashable, np.ndarray]], dims: tuple
    ) -> None:
        from sparse import COO, GCXS

        if isinstance(idx, pd.MultiIndex):
            coords = np.stack([np.asarray(code) for code in idx.codes], axis=0)
            is_sorted = idx.is_monotonic_increasing
            shape = tuple(lev.size for lev in idx.levels)
        else:
            coords = np.arange(idx.size).reshape(1, -1)
            is_sorted = True
            shape = (idx.size,)

        if not is_sorted:
            raise ValueError("index must be sorted")

        template_data = COO(
            coords,
            np.zeros(len(idx)),
            shape,
            has_duplicates=False,
            sorted=is_sorted,
            fill_value=np.nan,
        )
        template_gcxs = GCXS(template_data, compressed_axes=(0,))

        for name, values in arrays:
            # In virtually all real use cases, the sparse array will now have
            # missing values and needs a fill_value. For consistency, don't
            # special case the rare exceptions (e.g., dtype=int without a
            # MultiIndex).
            dtype, fill_value = dtypes.maybe_promote(values.dtype)
            values = np.asarray(values, dtype=dtype)

            data = GCXS(
                (values, template_gcxs.indices, template_gcxs.indptr),
                shape=template_gcxs.shape,
                compressed_axes=(0,),
                prune=False,
            )
            self[name] = (dims, data)

    def __init__(self, *args, **kwargs):
        pass # init for superclass happens inside __new__

    @property
    def CASEID(self):
        result = self.attrs.get(_CASEID, None)
        if result is None:
            warnings.warn("no defined CASEID")
            return _CASEID
        return result

    @CASEID.setter
    def CASEID(self, dim_name):
        if dim_name not in self.dims:
            raise ValueError(f"{dim_name} not in dims")
        self.attrs[_CASEID] = dim_name

    @property
    def ALTID(self):
        result = self.attrs.get(_ALTID, None)
        if result is None:
            warnings.warn("no defined ALTID")
            return _ALTID
        return result

    @ALTID.setter
    def ALTID(self, dim_name):
        self.attrs[_ALTID] = dim_name

    @property
    def alts_mapping(self):
        """Dict[int,str] : Mapping of alternative codes to names"""
        a = self.coords[self.ALTID]
        if 'alt_names' in a.coords:
            return dict(zip(a.values, a.coords['alt_names'].values))
        else:
            return dict(zip(a.values, a.values))

    @property
    def n_cases(self):
        try:
            return self.dims[self.CASEID]
        except KeyError:
            logging.getLogger().error(f"missing {self.CASEID!r} among dims {self.dims}")
            raise

    @property
    def n_alts(self):
        if self.ALTID in self.dims:
            return self.dims[self.ALTID]
        if 'n_alts' in self.attrs:
            return self.attrs['n_alts']
        raise ValueError('no n_alts set')

    @property
    def CASEALT(self):
        """str : The _casealt_ dimension of this Dataset, if defined."""
        result = self.attrs.get(_CASEALT, None)
        return result

    @CASEALT.setter
    def CASEALT(self, dim_name):
        self.attrs[_CASEALT] = dim_name

    @property
    def ALTIDX(self):
        """str : The _alt_idx_ dimension of this Dataset, if defined."""
        result = self.attrs.get(_ALTIDX, None)
        return result

    @ALTIDX.setter
    def ALTIDX(self, dim_name):
        self.attrs[_ALTIDX] = dim_name

    @property
    def CASEPTR(self):
        """str : The _caseptr_ dimension of this Dataset, if defined."""
        result = self.attrs.get(_CASEPTR, None)
        return result

    @CASEPTR.setter
    def CASEPTR(self, dim_name):
        self.attrs[_CASEPTR] = dim_name

    def _repr_html_(self):
        html = super()._repr_html_()
        html = html.replace("sharrow.Dataset", "larch.Dataset")
        return html

    def __repr__(self):
        r = super().__repr__()
        r = r.replace("sharrow.Dataset", "larch.Dataset")
        return r

    def __getattr__(self, name):
        # change attrs from xr.DataArray to larch.DataArray
        result = super().__getattr__(name)
        if isinstance(result, xr.DataArray) and not isinstance(result, DataArray):
            c = self.attrs.get(_CASEID, None)
            a = self.attrs.get(_ALTID, None)
            if c is not None:
                result.attrs[_CASEID] = c
            if a is not None:
                result.attrs[_ALTID] = a
            result.__class__ = DataArray
        return result

    def __getitem__(self, name):
        # change items from xr.DataArray to larch.DataArray
        result = super().__getitem__(name)
        if isinstance(result, xr.DataArray) and not isinstance(result, DataArray):
            c = self.attrs.get(_CASEID, None)
            a = self.attrs.get(_ALTID, None)
            if c is not None:
                result.attrs[_CASEID] = c
            if a is not None:
                result.attrs[_ALTID] = a
            result.__class__ = DataArray
        return result

    def get_expr(self, expression):
        """
        Access or evaluate an expression.

        Parameters
        ----------
        expression : str

        Returns
        -------
        DataArray
        """
        try:
            result = self[expression]
        except (KeyError, IndexError):
            try:
                self._flow_library
            except AttributeError:
                flow = self.setup_flow({expression: expression})
            else:
                if expression in self._flow_library:
                    flow = self._flow_library[expression]
                else:
                    flow = self.setup_flow({expression: expression})
                    self._flow_library[expression] = flow
            if not flow.tree.root_dataset is self:
                flow.tree = self.as_tree()
            result = flow.load_dataarray().isel(expressions=0)
        return result


    def to_arrays(self, graph, float_dtype=np.float64):
        from .numba.data_arrays import DataArrays
        from .numba.cascading import array_av_cascade, array_ch_cascade

        if 'co' in self:
            co = self['co'].values.astype(float_dtype)
        else:
            co = np.empty( (self.n_cases, 0), dtype=float_dtype)

        if 'ca' in self:
            ca = self['ca'].values.astype(float_dtype)
        else:
            ca = np.empty( (self.n_cases, self.n_alts, 0), dtype=float_dtype)

        if 'ce_data' in self:
            ce_data = self['ce_data'].values.astype(float_dtype)
        else:
            ce_data = np.empty( (0, 0), dtype=float_dtype)

        if self.ALTIDX is not None:
            ce_altidx = self[self.ALTIDX].values
        else:
            ce_altidx = np.empty( (0), dtype=np.int16)

        if self.CASEPTR is not None:
            ce_caseptr = np.lib.stride_tricks.sliding_window_view(
                self[self.CASEPTR].values, 2
            )
        else:
            ce_caseptr = np.empty( (self.n_cases, 0), dtype=np.int16)

        if 'wt' in self:
            wt = self['wt'].values.astype(float_dtype)
        else:
            wt = np.ones(self.n_cases, dtype=float_dtype)

        if 'ch' in self:
            ch = array_ch_cascade(self['ch'].values, graph, dtype=float_dtype)
        else:
            ch = np.zeros([self.n_cases, len(graph)], dtype=float_dtype)

        if 'av' in self:
            av = array_av_cascade(self['av'].values, graph)
        else:
            av = np.ones([self.n_cases, len(graph)], dtype=np.int8)

        return DataArrays(
            ch, av, wt, co, ca, ce_data, ce_altidx, ce_caseptr
        )

    def validate_format(self):
        error_msgs = []
        warn_msgs = []
        if self.CASEID not in self.dims:
            error_msgs.append(
                f"- There is no dimensions named `{self.CASEID}`. "
            )
        if self.ALTID not in self.dims:
            warn_msgs.append(
                f"- There is no dimensions named `_altid_`. "
            )
        msgs = []
        if error_msgs:
            msgs.append("ERRORS:")
            msgs.extend(error_msgs)
        if warn_msgs:
            msgs.append("WARNINGS:")
            msgs.extend(warn_msgs)
        return msgs

    def query_cases(self, query, parser="pandas", engine=None):
        """
        Return a new dataset with each array indexed along the CASEID dimension.

        The indexers are given as strings containing Python expressions to be
        evaluated against the data variables in the dataset.

        Parameters
        ----------
        query : str
            Python expressions to be evaluated against the data variables
            in the dataset. The expressions will be evaluated using the pandas
            eval() function, and can contain any valid Python expressions but cannot
            contain any Python statements.
        parser : {"pandas", "python"}, default: "pandas"
            The parser to use to construct the syntax tree from the expression.
            The default of 'pandas' parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the 'python'
            parser to retain strict Python semantics.
        engine : {"python", "numexpr", None}, default: None
            The engine used to evaluate the expression. Supported engines are:

            - None: tries to use numexpr, falls back to python
            - "numexpr": evaluates expressions using numexpr
            - "python": performs operations as if you had eval’d in top level python

        Returns
        -------
        obj : Dataset
            A new Dataset with the same contents as this dataset, except each
            array is indexed by the results of the query on the CASEID dimension.

        See Also
        --------
        Dataset.isel
        pandas.eval
        """
        return self.query({self.CASEID: query}, parser=parser, engine=engine)

    def dissolve_coords(self, dim, others=None):
        d = self.reset_index(dim)
        a = d[f"{dim}_"]
        mapper = dict((j, i) for (i, j) in enumerate(a.to_series()))
        mapper_f = np.vectorize(mapper.get)
        if others is None:
            others = []
        if isinstance(others, str):
            others = [others]
        for other in others:
            d[other] = xr.apply_ufunc(mapper_f, d[other])
        return d

    def dissolve_zero_variance(self, dim='<ALTID>', inplace=False):
        """
        Dissolve dimension on variables where it has no variance.

        This method is convenient to convert variables that have
        been loaded as |idca| or |idce| format into |idco| format where
        appropriate.

        Parameters
        ----------
        dim : str, optional
            The name of the dimension to potentially dissolve.
        inplace : bool, default False
            Whether to dissolve variables in-place.

        Returns
        -------
        Dataset
        """
        if dim == '<ALTID>':
            dim = self.ALTID
        if inplace:
            obj = self
        else:
            obj = self.copy()
        for k in obj.variables:
            if obj[k].dtype.kind in {'U', 'S', 'O'}:
                continue
            if dim in obj[k].dims:
                try:
                    dissolve = (obj[k].std(dim=dim).max() < 1e-10)
                except TypeError:
                    pass
                else:
                    if dissolve:
                        obj[k] = obj[k].min(dim=dim)
            elif obj[k].dims == (self.CASEALT,):
                proposal, flag = ce_dissolve_zero_variance(obj[k].values, obj[obj.CASEPTR].values)
                if flag == 0:
                    obj = obj.assign({k: DataArray(proposal, dims=(obj.CASEID))})
        return obj

    def set_dtypes(self, dtypes, inplace=False, on_error='warn'):
        """
        Set the dtypes for the variables in this Dataset.

        Parameters
        ----------
        dtypes : Mapping or DataFrame
            Mapping of names to dtypes, or a DataFrame to infer such a
            mapping.
        inplace : bool, default False
            Whether to convert dtypes inplace.
        on_error : {'warn', 'raise', 'ignore'}
            What to do when a type conversion triggers an error.

        Returns
        -------
        Dataset
        """
        if isinstance(dtypes, pd.DataFrame):
            dtypes = dtypes.dtypes
        if inplace:
            obj = self
        else:
            obj = self.copy()
        for k in obj:
            if k not in dtypes:
                continue
            try:
                obj[k] = obj[k].astype(dtypes[k])
            except Exception as err:
                if on_error == 'warn':
                    warnings.warn(f"{err!r} on converting {k}")
                elif on_error == 'raise':
                    raise
        return obj

    def set_altnames(self, altnames, inplace=False):
        """
        Set the alternative names for this Dataset.

        Parameters
        ----------
        altnames : Mapping or array-like
            A mapping of (integer) codes to names, or an array or names
            of the same length and order as the alternatives already
            defined in this Dataset.

        Returns
        -------
        Dataset
        """
        if inplace:
            obj = self
        else:
            obj = self.copy()
        if isinstance(altnames, Mapping):
            names = xr.DataArray(
                [altnames.get(i, None) for i in obj[obj.ALTID].values],
                dims=obj.ALTID,
            )
        elif isinstance(altnames, DataArray):
            names = altnames
        else:
            names = xr.DataArray(
                np.asarray(altnames),
                dims=obj.ALTID,
            )
        obj.coords['altnames'] = names
        return obj

    @classmethod
    def from_idca(cls, df, crack=True, altnames=None, avail='_avail_', fill_missing=None):
        """
        Construct a Dataset from an idca-format DataFrame.

        This method loads the data as dense arrays.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default True
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        avail : str, default '_avail_'
            When the imported data is in idce format (i.e. sparse) then
            an availability indicator is computed and given this name.
        fill_missing : scalar or Mapping, optional
            Fill values to use for missing values when imported data is
            in idce format (i.e. sparse).  Give a single value to use
            globally, or a mapping of {variable: value} or {dtype: value}.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.from_idce : Construct a Dataset from a sparse idca-format DataFrame.

        """
        if df.index.nlevels != 2:
            raise ValueError("source idca dataframe must have a two "
                             "level MultiIndex giving case and alt id's")
        caseidname, altidname = df.index.names

        # check altids are integers, if they are not then fix it
        if df.index.levels[1].dtype.kind != 'i':
            if altnames is None:
                altnames = df.index.levels[1]
                df.index = df.index.set_levels(np.arange(1, len(altnames) + 1), level=1)
            else:
                new_index = df.index.get_level_values(1).astype(pd.CategoricalDtype(altnames))
                df.index = df.index.set_codes(new_index.codes, level=1).set_levels(np.arange(1, len(altnames) + 1), level=1)

        ds = cls.construct(df, caseid=caseidname, alts=altidname)
        if crack:
            ds = ds.dissolve_zero_variance()
        ds = ds.set_dtypes(df)
        if altnames is not None:
            ds = ds.set_altnames(altnames)
        if avail not in ds and len(df) < ds.n_cases * ds.n_alts:
            av = DataArray.from_series(pd.Series(1, index=df.index)).fillna(0).astype(np.int8)
            ds[avail] = av
            if fill_missing is not None:
                if isinstance(fill_missing, Mapping):
                    for k, i in ds.items():
                        if ds.ALTID not in i.dims:
                            continue
                        if k not in fill_missing and i.dtype not in fill_missing:
                            continue
                        filler = fill_missing.get(k, fill_missing[i.dtype])
                        ds[k] = i.where(ds['_avail_']!=0, filler)
                else:
                    for k, i in ds.items():
                        if ds.ALTID not in i.dims:
                            continue
                        ds[k] = i.where(ds['_avail_']!=0, fill_missing)
        return ds

    @classmethod
    def from_idce(cls, df, crack=True, altnames=None, dim_name=None, alt_index='alt_idx', case_index=None, case_pointer=None):
        """
        Construct a Dataset from a sparse idca-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default False
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        dim_name : str, optional
            Name to apply to the sparse index dimension.
        alt_index : str, default 'alt_idx'
            Add the alt index (position) for each sparse data row as a
            coords array with this name.
        case_index : str, optional
            Add the case index (position) for each sparse data row as a
            coords array with this name. If not given, this array is not
            stored but it can still be reconstructed later from the case
            pointers.
        case_pointer : str, optional
            Use this name for the case_ptr dimension, overriding the
            default.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.from_idca : Construct a dense Dataset from a idca-format DataFrame.
        """
        if df.index.nlevels != 2:
            raise ValueError("source idce dataframe must have a two "
                             "level MultiIndex giving case and alt id's")
        caseidname, altidname = df.index.names
        caseidname = caseidname or _CASEID
        altidname = altidname or _ALTID
        dim_name = dim_name or _CASEALT
        case_pointer = case_pointer or _CASEPTR
        ds = cls.from_dataframe(df.reset_index(drop=True).rename_axis(index=dim_name))
        ds.coords[caseidname] = xr.DataArray(df.index.levels[0], dims=caseidname)
        ds.coords[altidname] = xr.DataArray(df.index.levels[1], dims=altidname)
        if case_index is not None:
            ds.coords[case_index] = xr.DataArray(df.index.codes[0], dims=dim_name)
        if alt_index is None:
            raise ValueError('alt_index cannot be None')
        ds.coords[alt_index] = xr.DataArray(df.index.codes[1], dims=dim_name)
        ds.coords[case_pointer] = xr.DataArray(
            np.where(np.diff(df.index.codes[0], prepend=np.nan, append=np.nan))[0],
            dims=case_pointer,
        )
        ds.attrs['_exclude_dims_'] = (caseidname, altidname, case_pointer)
        ds.attrs[_CASEID] = caseidname
        ds.attrs[_ALTID] = altidname
        ds.attrs[_CASEALT] = dim_name
        ds.attrs[_ALTIDX] = alt_index
        ds.attrs[_CASEPTR] = case_pointer
        ds = ds.drop_vars(dim_name)
        if crack:
            ds = ds.dissolve_zero_variance()
        ds = ds.set_dtypes(df)
        if altnames is not None:
            ds = ds.set_altnames(altnames)
        return ds

    @classmethod
    def from_idco(cls, df, alts=None):
        """
        Construct a Dataset from an idco-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idco-format DataFrame, with
            the caseid's in a single-level index,
        alts : Mapping or array-like, optional
            If given as a mapping, links alternative codes to names.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        if df.index.nlevels != 1:
            raise ValueError("source idco dataframe must have a one "
                             "level Index giving case id's")
        caseidname = df.index.name
        ds = cls.construct(df, caseid=caseidname, alts=alts)
        ds = ds.set_dtypes(df)
        return ds

    def as_tree(self, label='main', exclude_dims=()):
        """
        Convert this Dataset to a DataTree.

        For |idco| and |idca| datasets, the result will generally be a
        single-node tree.  For |idce| data, there will be

        Parameters
        ----------
        label : str, default 'main'
            Name to use for the root node in the tree.
        exclude_dims : Tuple[str], optional
            Exclude these dimensions, in addition to any
            dimensions listed in the `_exclude_dims_` attribute.

        Returns
        -------
        DataTree
        """
        if self.CASEPTR is not None:
            case_index = case_ptr_to_indexes(self.dims[self.CASEALT], self[self.CASEPTR].values)
            obj = self.assign({'_case_index_': DataArray(case_index, dims=(self.CASEALT))})
            tree = DataTree(**{label: obj.drop_dims(self.CASEID)})
            ds = obj.keep_dims(self.CASEID)
            ds.attrs.pop('_exclude_dims_', None)
            ds.attrs.pop('_caseptr_', None)
            ds.attrs.pop('_casealt_', None)
            ds.attrs.pop('_alt_idx_', None)
            tree.add_dataset(
                'idcoVars',
                ds,
                relationships=(
                    f"{label}._case_index_ -> idcoVars.{self.CASEID}"
                )
            )
        else:
            tree = DataTree(**{label: self})
        return tree

    def setup_flow(self, *args, **kwargs):
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        This method creates a new `DataTree` with only this Dataset as
        the root Dataset labeled `main`.  All other arguments are passed
        through to `DataTree.setup_flow`.

        Returns
        -------
        Flow
        """
        return self.as_tree().setup_flow(*args, **kwargs)

    def caseids(self):
        """
        Access the caseids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        return self.indexes[self.CASEID]

    def altids(self):
        """
        Access the altids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        return self.indexes[self.ALTID]


class DataTree(_sharrow_DataTree):

    DatasetType = Dataset

    def __init__(
        self,
        graph=None,
        root_node_name=None,
        extra_funcs=(),
        extra_vars=None,
        cache_dir=None,
        relationships=(),
        force_digitization=False,
        **kwargs,
    ):
        super().__init__(
            graph=graph,
            root_node_name=root_node_name,
            extra_funcs=extra_funcs,
            extra_vars=extra_vars,
            cache_dir=cache_dir,
            relationships=relationships,
            force_digitization=force_digitization,
            **kwargs,
        )
        dim_order = []
        c = self.root_dataset.attrs.get(_CASEID, None)
        if c is not None:
            dim_order.append(c)
        a = self.root_dataset.attrs.get(_ALTID, None)
        if a is not None:
            dim_order.append(a)
        self.dim_order = tuple(dim_order)

    def idco_subtree(self):
        if 'idcoVars' in self.subspaces:
            return self.subspaces['idcoVars'].as_tree()
        return self.drop_dims(self.ALTID, ignore_missing_dims=True)

    @property
    def CASEID(self):
        """str : The _caseid_ dimension of the root Dataset."""
        result = self.root_dataset.attrs.get(_CASEID, None)
        if result is None:
            warnings.warn("no defined CASEID")
            return _CASEID
        return result

    @property
    def ALTID(self):
        """str : The _altid_ dimension of the root Dataset."""
        result = self.root_dataset.attrs.get(_ALTID, None)
        if result is None:
            warnings.warn("no defined ALTID")
            return _ALTID
        return result

    @property
    def CASEALT(self):
        """str : The _casealt_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_CASEALT, None)
        return result

    @property
    def ALTIDX(self):
        """str : The _alt_idx_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_ALTIDX, None)
        return result

    @property
    def CASEPTR(self):
        """str : The _caseptr_ dimension of the root Dataset, if defined."""
        result = self.root_dataset.attrs.get(_CASEPTR, None)
        return result

    @property
    def n_cases(self):
        """int : The size of the _caseid_ dimension of the root Dataset."""
        return self.root_dataset.dims[self.CASEID]

    @property
    def n_alts(self):
        """int : The size of the _altid_ dimension of the root Dataset."""
        return self.root_dataset.dims[self.ALTID]

    def query_cases(self, *args, **kwargs):
        """
        Return a new DataTree, with a query filter applied to the root Dataset.

        Parameters
        ----------
        query : str
            Python expressions to be evaluated against the data variables
            in the root dataset. The expressions will be evaluated using the pandas
            eval() function, and can contain any valid Python expressions but cannot
            contain any Python statements.
        parser : {"pandas", "python"}, default: "pandas"
            The parser to use to construct the syntax tree from the expression.
            The default of 'pandas' parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the 'python'
            parser to retain strict Python semantics.
        engine : {"python", "numexpr", None}, default: None
            The engine used to evaluate the expression. Supported engines are:

            - None: tries to use numexpr, falls back to python
            - "numexpr": evaluates expressions using numexpr
            - "python": performs operations as if you had eval’d in top level python

        Returns
        -------
        DataTree
            A new DataTree with the same contents as this DataTree, except each
            array of the root Dataset is indexed by the results of the query on
            the CASEID dimension.

        See Also
        --------
        Dataset.query_cases
        """
        obj = self.copy()
        obj.root_dataset = obj.root_dataset.query_cases(*args, **kwargs)
        return obj

    def caseids(self):
        """
        Access the caseids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        try:
            return self.root_dataset.indexes[self.CASEID]
        except KeyError:
            for k, v in self.subspaces.items():
                if self.CASEID in v.indexes:
                    return v.indexes[self.CASEID]
            raise

    def altids(self):
        """
        Access the altids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        try:
            return self.root_dataset.indexes[self.ALTID]
        except KeyError:
            for k, v in self.subspaces.items():
                if self.ALTID in v.indexes:
                    return v.indexes[self.ALTID]
            raise

    def setup_flow(self, *args, **kwargs):
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        Parameters
        ----------
        definition_spec : Dict[str,str]
            Gives the names and expressions that define the variables to
            create in this new `Flow`.
        cache_dir : Path-like, optional
            A location to write out generated python and numba code. If not
            provided, a unique temporary directory is created.
        name : str, optional
            The name of this Flow used for writing out cached files. If not
            provided, a unique name is generated. If `cache_dir` is given,
            be sure to avoid name conflicts with other flow's in the same
            directory.
        dtype : str, default "float32"
            The name of the numpy dtype that will be used for the output.
        boundscheck : bool, default False
            If True, boundscheck enables bounds checking for array indices, and
            out of bounds accesses will raise IndexError. The default is to not
            do bounds checking, which is faster but can produce garbage results
            or segfaults if there are problems, so try turning this on for
            debugging if you are getting unexplained errors or crashes.
        error_model : {'numpy', 'python'}, default 'numpy'
            The error_model option controls the divide-by-zero behavior. Setting
            it to ‘python’ causes divide-by-zero to raise exception like
            CPython. Setting it to ‘numpy’ causes divide-by-zero to set the
            result to +/-inf or nan.
        nopython : bool, default True
            Compile using numba's `nopython` mode.  Provided for debugging only,
            as there's little point in turning this off for production code, as
            all the speed benefits of sharrow will be lost.
        fastmath : bool, default True
            If true, fastmath enables the use of "fast" floating point transforms,
            which can improve performance but can result in tiny distortions in
            results.  See numba docs for details.
        parallel : bool, default True
            Enable or disable parallel computation for certain functions.
        readme : str, optional
            A string to inject as a comment at the top of the flow Python file.
        flow_library : Mapping[str,Flow], optional
            An in-memory cache of precompiled Flow objects.  Using this can result
            in performance improvements when repeatedly using the same definitions.
        extra_hash_data : Tuple[Hashable], optional
            Additional data used for generating the flow hash.  Useful to prevent
            conflicts when using a flow_library with multiple similar flows.
        write_hash_audit : bool, default True
            Writes a hash audit log into a comment in the flow Python file, for
            debugging purposes.
        hashing_level : int, default 1
            Level of detail to write into flow hashes.  Increase detail to avoid
            hash conflicts for similar flows.  Level 2 adds information about
            names used in expressions and digital encodings to the flow hash,
            which prevents conflicts but requires more pre-computation to generate
            the hash.
        dim_exclude : Collection[str], optional
            Exclude these root dataset dimensions from this flow.

        Returns
        -------
        Flow
        """
        if 'dim_exclude' not in kwargs:
            if '_exclude_dims_' in self.root_dataset.attrs:
                kwargs['dim_exclude'] = self.root_dataset.attrs['_exclude_dims_']
        return super().setup_flow(*args, **kwargs)


def merge(
        objects,
        compat= "no_conflicts",
        join= "outer",
        fill_value = dtypes.NA,
        combine_attrs = "override",
        *,
        caseid=None,
        alts=None,
):
    """
    Merge any number of xarray objects into a single larch.Dataset as variables.

    Parameters
    ----------
    objects : iterable of Dataset or iterable of DataArray or iterable of dict-like
        Merge together all variables from these objects. If any of them are
        DataArray objects, they must have a name.

    compat : {"identical", "equals", "broadcast_equals", "no_conflicts", "override"}, optional
        String indicating how to compare variables of the same name for
        potential conflicts:
        - "broadcast_equals": all values must be equal when variables are
          broadcast against each other to ensure common dimensions.
        - "equals": all values and dimensions must be the same.
        - "identical": all values, dimensions and attributes must be the
          same.
        - "no_conflicts": only values which are not null in both datasets
          must be equal. The returned dataset then contains the combination
          of all non-null values.
        - "override": skip comparing and pick variable from first dataset

    join : {"outer", "inner", "left", "right", "exact"}, optional
        String indicating how to combine differing indexes in objects.
        - "outer": use the union of object indexes
        - "inner": use the intersection of object indexes
        - "left": use indexes from the first object with each dimension
        - "right": use indexes from the last object with each dimension
        - "exact": instead of aligning, raise `ValueError` when indexes to be
          aligned are not equal
        - "override": if indexes are of same size, rewrite indexes to be
          those of the first object with that dimension. Indexes for the same
          dimension must have the same size in all objects.

    fill_value : scalar or dict-like, optional
        Value to use for newly missing values. If a dict-like, maps
        variable names to fill values. Use a data array's name to
        refer to its values.

    combine_attrs : {"drop", "identical", "no_conflicts", "drop_conflicts", \
                    "override"} or callable, default: "override"
        A callable or a string indicating how to combine attrs of the objects being
        merged:
        - "drop": empty attrs on returned Dataset.
        - "identical": all attrs must be the same on every object.
        - "no_conflicts": attrs from all objects are combined, any that have
          the same name must also have the same value.
        - "drop_conflicts": attrs from all objects are combined, any that have
          the same name but different values are dropped.
        - "override": skip comparing and copy attrs from the first dataset to
          the result.
        If a callable, it must expect a sequence of ``attrs`` dicts and a context object
        as its only parameters.

    caseid : str, optional, keyword only
        This named dimension will be marked as the '_caseid_' dimension.

    alts : str or Mapping or array-like, keyword only
        If given as a str, this named dimension will be marked as the
        '_altid_' dimension.  Otherwise, give a Mapping that defines
        alternative names and (integer) codes or an array of codes.

    Returns
    -------
    Dataset
        Dataset with combined variables from each object.
    """
    return Dataset(
        xr.merge(
            objects,
            compat="no_conflicts",
            join="outer",
            fill_value=dtypes.NA,
            combine_attrs="override",
        ),
        caseid=caseid,
        alts=alts,
    )


def choice_avail_summary(dataset, graph=None, availability_co_vars=None):
    """
    Generate a summary of choice and availability statistics.

    Parameters
    ----------
    dataset : Dataset
        The loaded dataset to summarize, which should have
        `ch` and `av` variables.
    graph : NestingTree, optional
        The nesting graph.
    availability_co_vars : dict, optional
        Also attach the definition of the availability conditions.

    Returns
    -------
    pandas.DataFrame
    """
    if graph is None:
        if 'ch' in dataset:
            ch_ = np.asarray(dataset['ch'].copy())
        else:
            ch_ = None
        av_ = np.asarray(dataset.get('av'))
    else:
        from .numba.cascading import array_av_cascade, array_ch_cascade

        ch_ = array_ch_cascade(dataset.get('ch'), graph)
        av_ = array_av_cascade(dataset.get('av'), graph)

    if ch_ is not None:
        ch = ch_.sum(0)
    else:
        ch = None

    if av_ is not None:
        av = av_.sum(0)
    else:
        av = None

    arr_wt = dataset.get('wt')
    if arr_wt is not None:
        if ch_ is not None:
            ch_w = pd.Series((ch_ * arr_wt.values).sum(0))
        else:
            ch_w = None
        if av_ is not None:
            av_w = pd.Series((av_ * arr_wt.values).sum(0))
        else:
            av_w = None
        show_wt = np.any(ch != ch_w)
    else:
        ch_w = ch
        av_w = av
        show_wt = False

    if av_ is not None:
        ch_[av_ > 0] = 0
    if ch_ is not None and ch_.sum() > 0:
        ch_but_not_av = ch_.sum(0)
        if arr_wt is not None:
            ch_but_not_av_w = pd.Series((ch_ * arr_wt.values).sum(0), index=ch_.columns)
        else:
            ch_but_not_av_w = ch_but_not_av
    else:
        ch_but_not_av = None
        ch_but_not_av_w = None

    from collections import OrderedDict
    od = OrderedDict()

    if graph is not None:
        idx = graph.standard_sort
        od['name'] = pd.Series(graph.standard_sort_names, index=idx)
    else:
        idx = dataset.altids()
        if dataset.get('alt_names') is not None:
            od['name'] = pd.Series(dataset.get('alt_names'), index=idx)

    if show_wt:
        od['chosen weighted'] = pd.Series(ch_w, index=idx)
        od['chosen unweighted'] = pd.Series(ch, index=idx)
        od['available weighted'] = pd.Series(av_w, index=idx)
        od['available unweighted'] = pd.Series(av, index=idx)
    else:
        od['chosen'] = pd.Series(ch, index=idx)
        od['available'] = pd.Series(av, index=idx)
    if ch_but_not_av is not None:
        if show_wt:
            od['chosen but not available weighted'] = pd.Series(ch_but_not_av_w, index=idx)
            od['chosen but not available unweighted'] = pd.Series(ch_but_not_av, index=idx)
        else:
            od['chosen but not available'] = pd.Series(ch_but_not_av, index=idx)

    if availability_co_vars is not None:
        od['availability condition'] = pd.Series(
            availability_co_vars.values(),
            index=availability_co_vars.keys(),
        )

    result = pd.DataFrame.from_dict(od)
    if graph is not None:
        totals = result.loc[graph.root_id, :]
        result.drop(index=graph.root_id, inplace=True)
    else:
        totals = result.sum()

    for tot in (
            'chosen',
            'chosen weighted',
            'chosen unweighted',
            'chosen but not available',
            'chosen but not available weighted',
            'chosen but not available unweighted',
            'chosen thus available',
            'not available so not chosen'
    ):
        if tot in totals:
            result.loc['< Total All Alternatives >', tot] = totals[tot]

    result.loc['< Total All Alternatives >', pd.isnull(result.loc['< Total All Alternatives >', :])] = ""
    result.drop('_root_', errors='ignore', inplace=True)

    if 'availability condition' in result:
        result['availability condition'] = result['availability condition'].fillna('')

    for i in (
            'chosen',
            'chosen but not available',
            'chosen thus available',
            'not available so not chosen',
    ):
        if i in result.columns and all(result[i] == result[i].astype(int)):
            result[i] = result[i].astype(int)

    for i in (
            'available',
    ):
        if i in result.columns:
            j = result.columns.get_loc(i)
            if all(result.iloc[:-1,j] == result.iloc[:-1,j].astype(int)):
                result.iloc[:-1,j] = result.iloc[:-1,j].astype(int)

    return result



@nb.njit
def ce_dissolve_zero_variance(ce_data, ce_caseptr):
    """

    Parameters
    ----------
    ce_data : array-like, shape [n_casealts] one-dim only
    ce_altidx
    ce_caseptr
    n_alts

    Returns
    -------
    out : ndarray
    flag : int
        1 if variance was detected, 0 if no variance was found and
        the `out` array is valid.
    """
    failed = 0
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:,-1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0], )
    out = np.zeros(shape, dtype=ce_data.dtype)
    c = 0
    out[0] = ce_data[0]
    for row in range(ce_data.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
            out[c] = ce_data[row]
        else:
            if out[c] != ce_data[row]:
                failed = 1
                break
    return out, failed


@nb.njit
def case_ptr_to_indexes(n_casealts, case_ptrs):
    case_index = np.zeros(n_casealts, dtype=np.int64)
    for c in range(case_ptrs.shape[0]-1):
        case_index[case_ptrs[c]:case_ptrs[c + 1]] = c
    return case_index



def example(extract='m', estimate=False):
    import larch.numba as lx

    m = lx.Model()

    m.title = "swissmetro example 01 (simple logit)"

    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }

    m.choice_co_code = 'CHOICE'

    from larch.roles import P,X
    m.utility_co[1] = P("ASC_TRAIN")
    m.utility_co[2] = 0
    m.utility_co[3] = P("ASC_CAR")
    m.utility_co[1] += X("TRAIN_TT") * P("B_TIME")
    m.utility_co[2] += X("SM_TT") * P("B_TIME")
    m.utility_co[3] += X("CAR_TT") * P("B_TIME")
    m.utility_co[1] += X("TRAIN_CO*(GA==0)") * P("B_COST")
    m.utility_co[2] += X("SM_CO*(GA==0)") * P("B_COST")
    m.utility_co[3] += X("CAR_CO") * P("B_COST")

    m.ordering = [
        ("ASCs", 'ASC.*',),
        ("LOS", 'B_.*',),
    ]

    import pandas as pd
    raw_data = pd.read_csv(lx.example_file('swissmetro.csv.gz')).rename_axis(index='CASEID')
    raw_data.head()

    keep = raw_data.eval("PURPOSE in (1,3) and CHOICE != 0")
    selected_data = raw_data[keep]

    ds = lx.Dataset.from_idco(selected_data, alts={1:'Train', 2:'SM', 3:'Car'})

    ds

    m.datatree = ds
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(15)
    m.maximize_loglike(method='SLSQP')

    m.calculate_parameter_covariance()

    m.parameter_summary()

    # TEST
    assert m.parameter_summary().data.to_markdown() == """
    |                       |   Value |   Std Err |   t Stat | Signif   |   Null Value |
    |:----------------------|--------:|----------:|---------:|:---------|-------------:|
    | ('ASCs', 'ASC_CAR')   | -0.155  |  0.0432   |    -3.58 | ***      |            0 |
    | ('ASCs', 'ASC_TRAIN') | -0.701  |  0.0549   |   -12.78 | ***      |            0 |
    | ('LOS', 'B_COST')     | -0.0108 |  0.000518 |   -20.91 | ***      |            0 |
    | ('LOS', 'B_TIME')     | -0.0128 |  0.000569 |   -22.46 | ***      |            0 |
    """[1:-1]
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]
